import objax
import jax.numpy as np
from .likelihoods import Gaussian, MultiLatentLikelihood
from .kernels import Independent, Separable
from jax import vmap
from jax.lax import scan
from jax.scipy.linalg import cholesky, cho_factor, cho_solve
from jax.lib import xla_bridge
from .utils import (
    inv,
    inv_vmap,
    diag,
    solve,
    transpose,
    input_admin,
    compute_conditional_statistics,
    build_joint,
    set_z_stats,
    temporal_conditional,
    temporal_conditional_infinite_horizon,
    sum_natural_params_by_group,
    gaussian_expected_log_lik,
    compute_cavity,
    mvn_logpdf,
    mvn_logpdf_,
    pep_constant
)
from .ops import (
    gaussian_conditional,
    sparse_gaussian_conditional,
    sparse_conditional_post_to_data,
    kalman_filter,
    kalman_filter_pairs,
    kalman_filter_meanfield,
    kalman_filter_pairs_meanfield,
    kalman_filter_infinite_horizon,
    kalman_filter_infinite_horizon_pairs,
    rauch_tung_striebel_smoother,
    rauch_tung_striebel_smoother_meanfield,
    rauch_tung_striebel_smoother_infinite_horizon,
    process_noise_covariance,
    blocktensor_to_blockdiagmatrix,
    blockdiagmatrix_to_blocktensor
)
import math
from jax.config import config
config.update("jax_enable_x64", True)

LOG2PI = math.log(2 * math.pi)


class GaussianDistribution(objax.Module):
    """
    A small class defined to handle the fact that we often need access to both the mean / cov parameterisation
    of a Gaussian and its natural parameterisation.
    Important note: for simplicity we let nat2 = inv(cov) rather than nat2 = -0.5inv(cov). The latter is the proper
    natural parameter, but for Gaussian distributions we need not worry about the -0.5 (it cancels out anyway).
    """

    def __init__(self, mean, covariance):
        self.mean_ = objax.StateVar(mean)
        self.covariance_ = objax.StateVar(covariance)
        nat1, nat2 = self.reparametrise(mean, covariance)
        self.nat1_, self.nat2_ = objax.StateVar(nat1), objax.StateVar(nat2)

    def __call__(self):
        return self.mean, self.covariance

    @property
    def mean(self):
        return self.mean_.value

    @property
    def covariance(self):
        return self.covariance_.value

    @property
    def nat1(self):
        return self.nat1_.value

    @property
    def nat2(self):
        return self.nat2_.value

    @staticmethod
    def reparametrise(param1, param2):
        chol = cho_factor(param2, lower=True)
        reparam1 = cho_solve(chol, param1)
        reparam2 = cho_solve(chol, np.tile(np.eye(param2.shape[1]), [param2.shape[0], 1, 1]))
        return reparam1, reparam2

    def update_mean_cov(self, mean, covariance):
        self.mean_.value = mean
        self.covariance_.value = covariance
        self.nat1_.value, self.nat2_.value = self.reparametrise(mean, covariance)

    def update_nat_params(self, nat1, nat2):
        self.nat1_.value = nat1
        self.nat2_.value = nat2
        self.mean_.value, self.covariance_.value = self.reparametrise(nat1, nat2)


class BaseModel(objax.Module):
    """
    The parent model class: initialises all the common model features and implements shared methods
    TODO: move as much of the generic functionality as possible from this class to the inference class.
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 func_dim=1):
        super().__init__()
        if X.ndim < 2:
            X = X[:, None]
        if Y.ndim < 2:
            Y = Y[:, None]
        self.X = np.array(X)
        self.Y = np.array(Y)
        self.kernel = kernel
        self.likelihood = likelihood
        self.num_data = self.X.shape[0]  # number of data
        self.func_dim = func_dim  # number of latent dimensions
        self.obs_dim = Y.shape[1]  # dimensionality of the observations, Y
        if isinstance(self.kernel, Independent):
            pseudo_lik_size = self.func_dim  # the multi-latent case
        else:
            pseudo_lik_size = self.obs_dim
        self.pseudo_likelihood = GaussianDistribution(
            mean=np.zeros([self.num_data, pseudo_lik_size, 1]),
            covariance=1e2 * np.tile(np.eye(pseudo_lik_size), [self.num_data, 1, 1])
        )
        self.posterior_mean = objax.StateVar(np.zeros([self.num_data, self.func_dim, 1]))
        self.posterior_variance = objax.StateVar(np.tile(np.eye(self.func_dim), [self.num_data, 1, 1]))
        self.ind = np.arange(self.num_data)
        self.num_neighbours = np.ones(self.num_data)
        self.mask_y = np.isnan(self.Y).reshape(Y.shape[0], Y.shape[1])
        if self.func_dim == self.obs_dim:
            self.mask_pseudo_y = self.mask_y
        elif isinstance(self.likelihood, MultiLatentLikelihood):
            self.mask_pseudo_y = None
        else:
            self.mask_pseudo_y = np.tile(self.mask_y, [1, pseudo_lik_size])

    def __call__(self, X=None):
        if X is None:
            self.update_posterior()
        else:
            return self.predict(X)

    def prior_sample(self, num_samps=1, X=None, seed=0):
        raise NotImplementedError

    def update_posterior(self):
        raise NotImplementedError

    def compute_log_lik(self, pseudo_y=None, pseudo_var=None):
        """ Compute the log likelihood of the pseudo model, i.e. the log normaliser of the approximate posterior """
        raise NotImplementedError

    def predict(self, X, R=None):
        raise NotImplementedError

    def predict_y(self, X, R=None, cubature=None):
        """
        predict y at new test locations X
        TODO: check non-Gaussian likelihoods
        """
        mean_f, var_f = self.predict(X, R)
        mean_f = mean_f.reshape(mean_f.shape[0], -1, 1)
        if not isinstance(self.likelihood, MultiLatentLikelihood):
            var_f = var_f.reshape(var_f.shape[0], -1, 1)
        mean_y, var_y = vmap(self.likelihood.predict, (0, 0, None))(mean_f, var_f, cubature)
        return np.squeeze(mean_y), np.squeeze(var_y)

    def negative_log_predictive_density(self, X, Y, R=None, cubature=None):
        predict_mean, predict_var = self.predict(X, R)
        if Y.ndim < 2:
            Y = Y.reshape(-1, 1)
        if isinstance(self.likelihood, MultiLatentLikelihood):
            predict_mean = predict_mean[..., None]
        elif (predict_mean.ndim > 1) and (predict_mean.shape[1] != Y.shape[1]):  # multi-latent case
            predict_mean, predict_var = predict_mean[..., None], predict_var[..., None] * np.eye(predict_var.shape[1])
        else:
            predict_mean, predict_var = predict_mean.reshape(-1, 1, 1), predict_var.reshape(-1, 1, 1)
        log_density = vmap(self.likelihood.log_density, (0, 0, 0, None))(
            Y.reshape(predict_mean.shape[0], -1, 1),
            predict_mean,
            predict_var,
            cubature
        )
        return -np.nanmean(log_density)

    def group_natural_params(self, nat1, nat2, batch_ind=None):
        if (batch_ind is not None) and (batch_ind.shape[0] != self.num_data):
            nat1 = self.pseudo_likelihood.nat1.at[batch_ind].set(nat1)
            nat2 = self.pseudo_likelihood.nat2.at[batch_ind].set(nat2)
        return nat1, nat2

    def conditional_posterior_to_data(self, batch_ind=None, post_mean=None, post_cov=None):
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        ind = self.ind[batch_ind]
        if post_mean is None:
            post_mean = self.posterior_mean.value[ind]
        if post_cov is None:
            post_cov = self.posterior_variance.value[ind]
        return post_mean, post_cov

    def conditional_data_to_posterior(self, mean_f, cov_f):
        return mean_f, cov_f

    def expected_density_pseudo(self):
        expected_density = vmap(gaussian_expected_log_lik)(  # parallel operation
            self.pseudo_likelihood.mean,
            self.posterior_mean.value,
            self.posterior_variance.value,
            self.pseudo_likelihood.covariance,
            self.mask_pseudo_y
        )
        return np.sum(expected_density)

    def compute_full_pseudo_lik(self):
        return self.pseudo_likelihood()

    def compute_full_pseudo_nat(self, batch_ind):
        return self.pseudo_likelihood.nat1[batch_ind], self.pseudo_likelihood.nat2[batch_ind]

    def cavity_distribution(self, batch_ind=None, power=1.):
        """ Compute the power EP cavity for the given data points """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        nat1lik_full, nat2lik_full = self.compute_full_pseudo_nat(batch_ind)

        # then compute the cavity
        cavity_mean, cavity_cov = vmap(compute_cavity, [0, 0, 0, 0, None])(
            self.posterior_mean.value[batch_ind],
            self.posterior_variance.value[batch_ind],
            nat1lik_full,
            nat2lik_full,
            power
        )
        return cavity_mean, cavity_cov

    def compute_ep_energy_terms(self, batch_ind=None, power=1.):
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        cavity_mean, cavity_cov = self.cavity_distribution(batch_ind, power)
        pseudo_y, pseudo_var = self.compute_full_pseudo_lik()

        lel_pseudo = mvn_logpdf_(
            pseudo_y,
            cavity_mean,
            pseudo_var / power + cavity_cov,
            self.mask_pseudo_y
        )
        lel_pseudo += vmap(pep_constant, [0, None, 0])(pseudo_var, power, self.mask_pseudo_y)  # PEP constant

        lZ = self.compute_log_lik(pseudo_y, pseudo_var)
        return (cavity_mean, cavity_cov), lel_pseudo, lZ


class GaussianProcess(BaseModel):
    """
    A standard (kernel-based) GP model with prior of the form
        f(t) ~ GP(0,k(t,t'))
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y):
        if isinstance(kernel, Separable):
            func_dim = 1
        elif isinstance(kernel, Independent):
            func_dim = kernel.num_kernels
        else:
            func_dim = 1
        super().__init__(kernel, likelihood, X, Y, func_dim=func_dim)
        self.cov_block_diag_ind = np.array(
            np.kron(np.eye(self.num_data), np.ones([self.func_dim, self.func_dim])),
            dtype=bool
        )

    def update_posterior(self):
        """
        Compute the approximate posterior distribution using standard Gaussian identities
        """
        mean, covariance = gaussian_conditional(self.kernel,
                                                self.pseudo_likelihood.mean,
                                                self.pseudo_likelihood.covariance,
                                                self.X)
        self.posterior_mean.value = mean.reshape(self.num_data, self.func_dim, 1)
        # self.posterior_variance.value = np.diag(covariance).reshape(self.num_data, 1, 1)
        self.posterior_variance.value = covariance[self.cov_block_diag_ind].reshape(
            self.num_data, self.func_dim, self.func_dim
        )

    def compute_log_lik(self, pseudo_y=None, pseudo_var=None):
        """
        Compute the log marginal likelihood of the pseudo model, i.e. the log normaliser of the approximate posterior
        """
        dim = 1  # TODO: implement multivariate case
        # TODO: won't match MarkovGP for batching with missings or for multidim input

        if pseudo_y is None:
            pseudo_y = self.pseudo_likelihood.mean
            pseudo_var = self.pseudo_likelihood.covariance
        pseudo_y = pseudo_y
        pseudo_var = pseudo_var

        Knn = self.kernel(self.X, self.X)
        # Ky = Knn + np.diag(np.squeeze(pseudo_var))  # single-latent version
        pseudo_var_block_diag = blocktensor_to_blockdiagmatrix(pseudo_var)
        Ky = Knn + pseudo_var_block_diag  # multi-latent version

        # ---- compute the marginal likelihood, i.e. the normaliser, of the pseudo model ----
        # pseudo_y = diag(pseudo_y)
        pseudo_y = pseudo_y.reshape(-1, 1)
        Ly, low = cho_factor(Ky, lower=True)
        log_lik_pseudo = (
            - 0.5 * np.sum(pseudo_y.T @ cho_solve((Ly, low), pseudo_y))
            - np.sum(np.log(np.diag(Ly)))
            - 0.5 * pseudo_y.shape[0] * dim * LOG2PI
        )

        return log_lik_pseudo

    def predict(self, X, R=None):
        """
        predict f at new test locations X
        """
        if len(X.shape) < 2:
            X = X[:, None]
        mean, covariance = gaussian_conditional(self.kernel,
                                                self.pseudo_likelihood.mean,
                                                self.pseudo_likelihood.covariance,
                                                self.X,
                                                X)
        predict_mean = mean.reshape(X.shape[0], -1)
        predict_variance = blockdiagmatrix_to_blocktensor(covariance, predict_mean.shape[0], predict_mean.shape[1])
        # predict_variance = vmap(np.diag)(np.diag(covariance).reshape(X.shape[0], -1))  # just diagonal
        return np.squeeze(predict_mean), np.squeeze(predict_variance)

    def compute_kl(self):
        """
        KL divergence between the approximate posterior q(u) and the prior p(u)
        """
        # log int p(u) prod_n N(pseudo_y_n | u, pseudo_var_n) du
        log_lik_pseudo = self.compute_log_lik()
        # E_q[log N(pseudo_y_n | u, pseudo_var_n)]
        expected_density_pseudo = self.expected_density_pseudo()
        kl = expected_density_pseudo - log_lik_pseudo  # KL[approx_post || prior]
        return kl

    def prior_sample(self, num_samps=1, X=None, seed=0):
        """
        Sample from the model prior f~N(0,K) multiple times using a nested loop.
        :param num_samps: the number of samples to draw [scalar]
        :param X: the input locations at which to sample (defaults to training inputs) [N, dim]
        :param seed: the random seed for sampling
        :return:
            f_samples: the prior samples [num_samps, N, func_dim]
        """
        if X is None:
            X = self.X
        if X.ndim < 2:
            X = X[:, None]
        N = X.shape[0]
        m = np.zeros([num_samps, N, self.func_dim])
        K = self.kernel(X, X)
        chol_K = np.tile(cholesky(K + 1e-12 * np.eye(N), lower=True), [num_samps, 1, 1])
        gen = objax.random.Generator(seed)
        rand_norm_samples = objax.random.normal(shape=(num_samps, N, self.func_dim), generator=gen)
        f_samples = m + chol_K @ rand_norm_samples
        return f_samples


GP = GaussianProcess


class SparseGaussianProcess(GaussianProcess):
    """
    A standard (kernel-based) GP model with prior of the form
        f(t) ~ GP(0,k(t,t'))
    :param opt_z: flag whether to optimise the inducing inputs Z
    TODO: write test comparing to gpflow
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 Z,
                 opt_z=False):
        super().__init__(kernel, likelihood, X, Y)
        if Z.ndim < 2:
            Z = Z[:, None]
        if opt_z:
            self.Z = objax.TrainVar(np.array(Z))
        else:
            self.Z = objax.StateVar(np.array(Z))
        self.num_inducing = Z.shape[0]
        self.posterior_mean = objax.StateVar(np.zeros([self.num_inducing, self.func_dim, 1]))
        self.posterior_variance = objax.StateVar(np.tile(np.eye(self.func_dim), [self.num_inducing, 1, 1]))
        self.posterior_covariance = objax.StateVar(np.eye(self.func_dim * self.num_inducing))
        if self.func_dim != self.obs_dim:
            self.mask_pseudo_y = None  # TODO: fix mask for multi-latent case
        self.cov_block_diag_ind = np.array(
            np.kron(np.eye(self.num_inducing), np.ones([self.func_dim, self.func_dim])),
            dtype=bool
        )

    def update_posterior(self):
        """
        Compute the approximate posterior distribution using standard Gaussian identities
        """
        mean, covariance = sparse_gaussian_conditional(self.kernel,
                                                       self.pseudo_likelihood.nat1,
                                                       self.pseudo_likelihood.nat2,
                                                       self.X,
                                                       self.Z.value)
        self.posterior_mean.value = mean.reshape(self.num_inducing, self.func_dim, 1)
        # self.posterior_variance.value = np.diag(covariance).reshape(self.num_inducing, 1, 1)
        self.posterior_variance.value = covariance[self.cov_block_diag_ind].reshape(
            self.num_inducing, self.func_dim, self.func_dim
        )
        self.posterior_covariance.value = covariance.reshape(
            self.func_dim * self.num_inducing,
            self.func_dim * self.num_inducing
        )

    def compute_global_pseudo_lik(self):
        nat1lik_global, nat2lik_global = self.compute_global_pseudo_nat()
        pseudo_var_global = inv(nat2lik_global + 1e-12 * np.eye(nat2lik_global.shape[0]))
        pseudo_y_global = pseudo_var_global @ nat1lik_global
        return pseudo_y_global, pseudo_var_global

    def compute_global_pseudo_nat(self):
        """ The pseudo-likelihoods are currently stored as N Gaussians in f """
        Kuf = self.kernel(self.Z.value, self.X)  # only compute log lik for observed values
        Kuu = self.kernel(self.Z.value, self.Z.value)
        Wuf = solve(Kuu, Kuf)  # conditional mapping, Kuu^-1 Kuf

        nat1lik_global = Wuf @ self.pseudo_likelihood.nat1.reshape(-1, 1)
        nat2lik_block_diag = blocktensor_to_blockdiagmatrix(self.pseudo_likelihood.nat2)  # multi-latent version
        nat2lik_global = Wuf @ nat2lik_block_diag @ transpose(Wuf)
        return nat1lik_global, nat2lik_global

    def compute_full_pseudo_lik(self, batch_ind=None):
        nat1lik_full, nat2lik_full = self.compute_full_pseudo_nat(batch_ind)
        pseudo_var_full = inv_vmap(nat2lik_full + 1e-12 * np.eye(nat2lik_full.shape[1]))
        pseudo_y_full = pseudo_var_full @ nat1lik_full
        return pseudo_y_full, pseudo_var_full

    def compute_full_pseudo_nat(self, batch_ind=None):
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        Kuf = self.kernel(self.Z.value, self.X[batch_ind].reshape(-1, 1))  # only compute log lik for observed values
        Kuu = self.kernel(self.Z.value, self.Z.value)
        Nbatch = batch_ind.shape[0]
        Wuf = solve(Kuu, Kuf).T.reshape(Nbatch, self.func_dim, -1)  # conditional mapping, Kuu^-1 Kuf
        nat1lik_full = transpose(Wuf) @ self.pseudo_likelihood.nat1[batch_ind]
        nat2lik_full = transpose(Wuf) @ self.pseudo_likelihood.nat2[batch_ind] @ Wuf
        return nat1lik_full, nat2lik_full

    def compute_kl(self):
        """
        KL divergence between the approximate posterior q(u) and the prior p(u)
        """
        pseudo_y_full, pseudo_var_full = self.compute_global_pseudo_lik()

        # ---- compute the log marginal likelihood, i.e. the normaliser, of the pseudo model ----
        # log int p(u) prod_n N(pseudo_y_n | u, pseudo_var_n) du
        log_lik_pseudo = self.compute_log_lik(pseudo_y_full, pseudo_var_full)

        # E_q[log N(pseudo_y_n | u, pseudo_var_n)]
        expected_density_pseudo = gaussian_expected_log_lik(  # this term does not depend on the prior, use stored q(u)
            pseudo_y_full,
            self.posterior_mean.value.reshape(-1, 1),
            self.posterior_covariance.value,
            pseudo_var_full
        )

        kl = expected_density_pseudo - log_lik_pseudo  # KL[approx_post || prior]
        return kl

    def compute_log_lik(self, pseudo_y=None, pseudo_var=None):
        """ log int p(u) prod_n N(pseudo_y_n | u, pseudo_var_n) du """
        dim = 1.  # TODO: implement multivariate case
        Kuu = self.kernel(self.Z.value, self.Z.value)

        Ky = Kuu + pseudo_var
        Ly, low = cho_factor(Ky, lower=True)
        log_lik_pseudo = (  # this term depends on the prior
            - 0.5 * np.sum(pseudo_y.T @ cho_solve((Ly, low), pseudo_y))
            - np.sum(np.log(np.diag(Ly)))
            - 0.5 * pseudo_y.shape[0] * dim * LOG2PI
        )
        return log_lik_pseudo

    def predict(self, X, R=None):
        """
        predict at new test locations X
        """
        if len(X.shape) < 2:
            X = X[:, None]
        self.update_posterior()
        mean, covariance = sparse_conditional_post_to_data(self.kernel,
                                                           self.posterior_mean.value,
                                                           self.posterior_covariance.value,
                                                           X,
                                                           self.Z.value)
        predict_mean = mean.reshape(X.shape[0], -1)
        predict_variance = blockdiagmatrix_to_blocktensor(covariance, predict_mean.shape[0], predict_mean.shape[1])
        return predict_mean, predict_variance

    def conditional_posterior_to_data(self, batch_ind=None, post_mean=None, post_cov=None):
        """
        compute
        q(f) = int p(f | u) q(u) du
        where
        q(u) = N(u | post_mean, post_cov)
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        if post_mean is None:
            post_mean = self.posterior_mean.value
        if post_cov is None:
            post_cov = self.posterior_covariance.value

        ndim = post_cov.ndim
        Nbatch = batch_ind.shape[0]

        if ndim < 3:
            mean_f, cov_f = sparse_conditional_post_to_data(self.kernel,
                                                            post_mean,
                                                            post_cov,
                                                            self.X[batch_ind],
                                                            self.Z.value)
            var_f = blockdiagmatrix_to_blocktensor(cov_f, Nbatch, self.func_dim)
        else:  # in the PEP case we have one cavity per data point
            mean_f, var_f = vmap(sparse_conditional_post_to_data, [None, 0, 0, 0, None])(self.kernel,
                                                                                         post_mean,
                                                                                         post_cov,
                                                                                         self.X[batch_ind],
                                                                                         self.Z.value)
        return mean_f.reshape(Nbatch, self.func_dim, 1), var_f

    def cavity_distribution(self, batch_ind=None, power=1.):
        """ Compute the power EP cavity for the given data points """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        nat1lik_full, nat2lik_full = self.compute_full_pseudo_nat(batch_ind)

        # then compute the cavity
        cavity_mean, cavity_cov = vmap(compute_cavity, [None, None, 0, 0, None])(
            self.posterior_mean.value.reshape(-1, 1),
            self.posterior_covariance.value,
            nat1lik_full,
            nat2lik_full,
            power
        )
        return cavity_mean, cavity_cov

    def cavity_distribution_tied(self, batch_ind=None, power=1.):
        """ The "tied" version rather than the more accurate but much slower cavity computation above """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        N = batch_ind.shape[0]
        # Kuf = self.kernel(self.Z.value, self.X)  # only compute log lik for observed values
        # Kuu = self.kernel(self.Z.value, self.Z.value)
        # Wuf = solve(Kuu, Kuf)  # conditional mapping, Kuu^-1 Kuf
        #
        # nat1lik_full = Wuf @ np.squeeze(self.pseudo_likelihood.nat1, axis=-1)
        # nat2lik_full = Wuf @ np.diag(np.squeeze(self.pseudo_likelihood.nat2)) @ transpose(Wuf)
        #
        # nat1cav = nat1lik_full * (self.num_inducing - power) / self.num_inducing
        # nat2cav = nat2lik_full * (self.num_inducing - power) / self.num_inducing
        #
        # cavity_cov = inv(nat2cav + 1e-12 * np.eye(Kuu.shape[0]))
        # cavity_mean = cavity_cov @ nat1cav
        nat1lik, nat2lik = self.compute_global_pseudo_nat()
        cavity_mean, cavity_cov = compute_cavity(
            self.posterior_mean.value.reshape(-1, 1),
            self.posterior_covariance.value,
            nat1lik,
            nat2lik,
            power / N
        )
        return np.tile(cavity_mean, [N, 1, 1]), np.tile(cavity_cov, [N, 1, 1])

    def compute_ep_energy_terms(self, batch_ind=None, power=1.):
        """ Here we use the "tied" version rather than the full cavity computation """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        N = batch_ind.shape[0]
        nat1lik, nat2lik = self.compute_global_pseudo_nat()
        pseudo_var = inv(nat2lik + 1e-12 * np.eye(nat2lik.shape[0]))
        pseudo_y = pseudo_var @ nat1lik

        post_nat2 = inv(self.posterior_covariance.value + 1e-8 * np.eye(self.posterior_covariance.value.shape[1]))
        post_nat1 = post_nat2 @ self.posterior_mean.value.reshape(-1, 1)
        global_cavity_cov = inv(post_nat2 - power * nat2lik)
        local_cavity_cov = inv(post_nat2 - power / N * nat2lik)
        global_cavity_mean = global_cavity_cov @ (post_nat1 - power * nat1lik)
        local_cavity_mean = local_cavity_cov @ (post_nat1 - power / N * nat1lik)

        lel_pseudo = mvn_logpdf(
            pseudo_y,
            global_cavity_mean,
            pseudo_var / power + global_cavity_cov
        )
        lel_pseudo += pep_constant(pseudo_var, power)  # add PEP constant to account for power in normaliser
        lZ = self.compute_log_lik(pseudo_y, pseudo_var)
        return (np.tile(local_cavity_mean, [N, 1, 1]), np.tile(local_cavity_cov, [N, 1, 1])), lel_pseudo, lZ


SparseGP = SparseGaussianProcess


class MarkovGaussianProcess(BaseModel):
    """
    The stochastic differential equation (SDE) form of a Gaussian process (GP) model.
    Implements methods for inference and learning using state space methods, i.e. Kalman filtering and smoothing.
    Constructs a linear time-invariant (LTI) stochastic differential equation (SDE) of the following form:
        dx(t)/dt = F x(t) + L w(t)
              yₙ ~ p(yₙ | f(t_n)=H x(t_n))
    where w(t) is a white noise process and where the state x(t) is Gaussian distributed with initial
    state distribution x(t)~𝓝(0,Pinf).
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 R=None,
                 parallel=None):
        if parallel is None:  # if using a GPU, then run the parallel filter
            parallel = xla_bridge.get_backend().platform == 'gpu'
        (X, Y, self.R, self.dt) = input_admin(X, Y, R)
        H = kernel.measurement_model()
        func_dim = H.shape[0]  # number of latent dimensions
        super().__init__(kernel, likelihood, X, Y, func_dim=func_dim)
        self.state_dim = self.kernel.stationary_covariance().shape[0]
        self.minf = np.zeros([self.state_dim, 1])  # stationary state mean
        self.spatio_temporal = np.any(~np.isnan(self.R))
        self.parallel = parallel
        if (self.func_dim != self.obs_dim) and self.spatio_temporal:
            self.mask_pseudo_y = None  # sparse spatio-temporal case, no mask required

    @staticmethod
    def filter(*args, **kwargs):
        return kalman_filter(*args, **kwargs)

    @staticmethod
    def smoother(*args, **kwargs):
        return rauch_tung_striebel_smoother(*args, **kwargs)

    @staticmethod
    def temporal_conditional(*args, **kwargs):
        return temporal_conditional(*args, **kwargs)

    def compute_full_pseudo_nat(self, batch_ind):
        if self.spatio_temporal:  # spatio-temporal case
            B, C = self.kernel.spatial_conditional(self.X[batch_ind], self.R[batch_ind])
            nat1lik_full = transpose(B) @ self.pseudo_likelihood.nat1[batch_ind]
            nat2lik_full = transpose(B) @ self.pseudo_likelihood.nat2[batch_ind] @ B
            return nat1lik_full, nat2lik_full
        else:  # temporal case
            return self.pseudo_likelihood.nat1[batch_ind], self.pseudo_likelihood.nat2[batch_ind]

    def compute_full_pseudo_lik(self):
        # TODO: running this 3 times per training loop is wasteful - store in memory?
        if self.spatio_temporal:  # spatio-temporal case
            B, C = self.kernel.spatial_conditional(self.X, self.R)
            # TODO: more efficient way to do this?
            nat1lik_full = transpose(B) @ self.pseudo_likelihood.nat1
            nat2lik_full = transpose(B) @ self.pseudo_likelihood.nat2 @ B
            pseudo_var_full = inv_vmap(nat2lik_full + 1e-12 * np.eye(nat2lik_full.shape[1]))  # <---------- bottleneck
            pseudo_y_full = pseudo_var_full @ nat1lik_full
            return pseudo_y_full, pseudo_var_full
        else:  # temporal case
            return self.pseudo_likelihood.mean, self.pseudo_likelihood.covariance

    def update_posterior(self):
        """
        Compute the posterior via filtering and smoothing
        """
        pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        log_lik, (filter_mean, filter_cov) = self.filter(self.dt,
                                                         self.kernel,
                                                         pseudo_y,
                                                         pseudo_var,
                                                         mask=self.mask_pseudo_y,
                                                         parallel=self.parallel)
        dt = np.concatenate([self.dt[1:], np.array([0.0])], axis=0)
        smoother_mean, smoother_cov, _ = self.smoother(dt,
                                                       self.kernel,
                                                       filter_mean,
                                                       filter_cov,
                                                       parallel=self.parallel)
        self.posterior_mean.value, self.posterior_variance.value = smoother_mean, smoother_cov

    def compute_kl(self):
        """
        KL[q()|p()]
        """
        pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        log_lik_pseudo = self.compute_log_lik(pseudo_y, pseudo_var)

        expected_density_pseudo = vmap(gaussian_expected_log_lik)(  # parallel operation
            pseudo_y,
            self.posterior_mean.value,
            self.posterior_variance.value,
            pseudo_var,
            self.mask_pseudo_y
        )

        kl = np.sum(expected_density_pseudo) - log_lik_pseudo  # KL[approx_post || prior]
        return kl

    def compute_log_lik(self, pseudo_y=None, pseudo_var=None):
        """
        int p(f) N(pseudo_y | f, pseudo_var) df
        """
        if pseudo_y is None:
            pseudo_y, pseudo_var = self.compute_full_pseudo_lik()

        log_lik_pseudo, (_, _) = self.filter(
            self.dt,
            self.kernel,
            pseudo_y,
            pseudo_var,
            mask=self.mask_pseudo_y,
            parallel=self.parallel
        )
        return log_lik_pseudo

    def conditional_posterior_to_data(self, batch_ind=None, post_mean=None, post_cov=None):
        """
        compute
        q(f) = int p(f | u) q(u) du = N(f | B post_mean, B post_cov B' + C)
        where
        q(u) = N(u | post_mean, post_cov)
        p(f | u) = N(f | Bu, C)
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        if post_mean is None:
            post_mean = self.posterior_mean.value[batch_ind]
        if post_cov is None:
            post_cov = self.posterior_variance.value[batch_ind]

        if self.spatio_temporal:
            B, C = self.kernel.spatial_conditional(self.X[batch_ind], self.R[batch_ind])
            mean_f = B @ post_mean
            cov_f = B @ post_cov @ transpose(B) + C
            return mean_f, cov_f
        else:
            return post_mean, post_cov

    def predict(self, X=None, R=None, pseudo_lik_params=None):
        """
        predict at new test locations X
        """
        if X is None:
            X = self.X
        elif len(X.shape) < 2:
            X = X[:, None]
        if R is None:
            R = X[:, 1:]
        X = X[:, :1]  # take only the temporal component

        if pseudo_lik_params is None:
            pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        else:
            pseudo_y, pseudo_var = pseudo_lik_params  # this deals with the posterior sampling case
        _, (filter_mean, filter_cov) = self.filter(self.dt,
                                                   self.kernel,
                                                   pseudo_y,
                                                   pseudo_var,
                                                   mask=self.mask_pseudo_y,  # mask has no effect here (loglik not used)
                                                   parallel=self.parallel)
        dt = np.concatenate([self.dt[1:], np.array([0.0])], axis=0)
        smoother_mean, smoother_cov, gain = self.smoother(dt,
                                                          self.kernel,
                                                          filter_mean,
                                                          filter_cov,
                                                          return_full=True,
                                                          parallel=self.parallel)

        # add dummy states at either edge
        inf = 1e10 * np.ones_like(self.X[0, :1])
        X_aug = np.block([[-inf], [self.X[:, :1]], [inf]])

        # predict the state distribution at the test time steps:
        state_mean, state_cov = self.temporal_conditional(X_aug, X, smoother_mean, smoother_cov, gain, self.kernel)
        # extract function values from the state:
        H = self.kernel.measurement_model()
        if self.spatio_temporal:
            # TODO: if R is fixed, only compute B, C once
            B, C = self.kernel.spatial_conditional(X, R)
            W = B @ H
            test_mean = W @ state_mean
            test_var = W @ state_cov @ transpose(W) + C
        else:
            test_mean, test_var = H @ state_mean, H @ state_cov @ transpose(H)

        # if np.squeeze(test_var).ndim > 2:  # deal with spatio-temporal case (discard spatial covariance)
        if self.spatio_temporal:  # deal with spatio-temporal case (discard spatial covariance)
            test_var = diag(test_var)
        return np.squeeze(test_mean), np.squeeze(test_var)

    def filter_energy(self):
        pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        _, (filter_mean, filter_cov) = self.filter(self.dt,
                                                   self.kernel,
                                                   pseudo_y,
                                                   pseudo_var,
                                                   mask=self.mask_pseudo_y,  # mask has no effect here (loglik not used)
                                                   parallel=self.parallel,
                                                   return_predict=True)
        H = self.kernel.measurement_model()
        mean = H @ filter_mean
        var = H @ filter_cov @ transpose(H)
        filter_energy = -np.sum(vmap(self.likelihood.log_density)(self.Y, mean, var))
        return filter_energy

    def prior_sample(self, num_samps=1, X=None, seed=0):
        """
        Sample from the model prior f~N(0,K) multiple times using a nested loop.
        :param num_samps: the number of samples to draw [scalar]
        :param X: the input locations at which to sample (defaults to training inputs) [N, 1]
        :param seed: the random seed for sampling
        :return:
            f_samples: the prior samples [num_samps, N, func_dim]
        """
        if X is None:
            dt = self.dt
        else:
            dt = np.concatenate([np.array([0.0]), np.diff(np.sort(X))])
        sd = self.state_dim
        H = self.kernel.measurement_model()
        Pinf = self.kernel.stationary_covariance()
        As = vmap(self.kernel.state_transition)(dt)
        Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)
        jitter = 1e-8 * np.eye(sd)
        f0 = np.zeros([dt.shape[0], self.func_dim, 1])

        def draw_full_sample(carry_, _):
            f_sample_i, i = carry_
            gen0 = objax.random.Generator(seed - 1 - i)
            m0 = cholesky(Pinf, lower=True) @ objax.random.normal(shape=(sd, 1), generator=gen0)

            def sample_one_time_step(carry, inputs):
                m, k = carry
                A, Q = inputs
                chol_Q = cholesky(Q + jitter, lower=True)  # <--- can be a bit unstable
                gen = objax.random.Generator(seed + i * k + k)
                q_samp = chol_Q @ objax.random.normal(shape=(sd, 1), generator=gen)
                m = A @ m + q_samp
                f = H @ m
                return (m, k+1), f

            (_, _), f_sample = scan(f=sample_one_time_step,
                                    init=(m0, 0),
                                    xs=(As, Qs))

            return (f_sample, i+1), f_sample

        (_, _), f_samples = scan(f=draw_full_sample,
                                 init=(f0, 0),
                                 xs=np.zeros(num_samps))

        return f_samples

    def posterior_sample(self, X=None, num_samps=1, seed=0):
        """
        Sample from the posterior at the test locations.
        Posterior sampling works by smoothing samples from the prior using the approximate Gaussian likelihood
        model given by the pseudo-likelihood, 𝓝(f|μ*,σ²*), computed during training.
         - draw samples (f*) from the prior
         - add Gaussian noise to the prior samples using auxillary model p(y*|f*) = 𝓝(y*|f*,σ²*)
         - smooth the samples by computing the posterior p(f*|y*)
         - posterior samples = prior samples + smoothed samples + posterior mean
                             = f* + E[p(f*|y*)] + E[p(f|y)]
        See Arnaud Doucet's note "A Note on Efficient Conditional Simulation of Gaussian Distributions" for details.
        :param X: the sampling input locations [N, 1]
        :param num_samps: the number of samples to draw [scalar]
        :param seed: the random seed for sampling
        :return:
            the posterior samples [N_test, num_samps]
        """
        if X is None:
            train_ind = np.arange(self.num_data)
            test_ind = train_ind
        else:
            if X.ndim < 2:
                X = X[:, None]
            X = np.concatenate([self.X, X])
            X, ind = np.unique(X, return_inverse=True)
            train_ind, test_ind = ind[:self.num_data], ind[self.num_data:]
        post_mean, _ = self.predict(X)
        prior_samp = self.prior_sample(X=X, num_samps=num_samps, seed=seed)  # sample at training locations
        lik_chol = np.tile(cholesky(self.pseudo_likelihood.covariance, lower=True), [num_samps, 1, 1, 1])
        gen = objax.random.Generator(seed)
        prior_samp_train = prior_samp[:, train_ind]
        prior_samp_y = prior_samp_train + lik_chol @ objax.random.normal(shape=prior_samp_train.shape, generator=gen)

        def smooth_prior_sample(i, prior_samp_y_i):
            smoothed_sample, _ = self.predict(X, pseudo_lik_params=(prior_samp_y_i, self.pseudo_likelihood.covariance))
            return i+1, smoothed_sample

        _, smoothed_samples = scan(f=smooth_prior_sample,
                                   init=0,
                                   xs=prior_samp_y)

        return (prior_samp[..., 0, 0] - smoothed_samples + post_mean[None])[:, test_ind]


MarkovGP = MarkovGaussianProcess


class SparseMarkovGaussianProcess(MarkovGaussianProcess):
    """
    A sparse Markovian GP.
    TODO: implement version with non-tied sites
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 Z,
                 R=None,
                 parallel=None):
        super().__init__(kernel, likelihood, X, Y, R=R, parallel=parallel)
        if Z.ndim < 2:
            Z = Z[:, None]
        Z = np.sort(Z, axis=0)
        inf = np.array([[1e10]])
        self.Z = objax.StateVar(np.concatenate([-inf, Z, inf], axis=0))
        self.dz = np.array(np.diff(self.Z.value[:, 0]))
        self.num_transitions = self.dz.shape[0]
        zeros = np.zeros([self.num_transitions, 2 * self.state_dim, 1])
        eyes = np.tile(np.eye(2 * self.state_dim), [self.num_transitions, 1, 1])

        # nat2 = 1e-8 * eyes

        # initialise to match MarkovGP / GP on first step (when Z=X):
        nat2 = (1e-8*eyes).at[:-1, self.state_dim, self.state_dim].set(1e-2)

        # initialise to match old implementation:
        # nat2 = (1 / 99) * eyes

        self.pseudo_likelihood = GaussianDistribution(
            zeros,
            inv_vmap(nat2)
        )
        self.posterior_mean = objax.StateVar(zeros)
        self.posterior_variance = objax.StateVar(eyes)
        self.mask_pseudo_y = None
        self.conditional_mean = None
        # TODO: if training Z this needs to be done at every training step (as well as sorting and computing dz)
        self.ind, self.num_neighbours = set_z_stats(self.X, self.Z.value)

    @staticmethod
    def filter(*args, **kwargs):
        return kalman_filter_pairs(*args, **kwargs)

    @staticmethod
    def smoother(*args, **kwargs):
        return rauch_tung_striebel_smoother(*args, **kwargs)

    def compute_full_pseudo_lik(self):
        return self.pseudo_likelihood()

    def update_posterior(self):
        """
        Compute the posterior via filtering and smoothing
        """
        log_lik, (filter_mean, filter_cov) = self.filter(self.dz,
                                                         self.kernel,
                                                         self.pseudo_likelihood.mean,
                                                         self.pseudo_likelihood.covariance,
                                                         parallel=self.parallel)
        dz = self.dz[1:]
        smoother_mean, smoother_cov, gain = self.smoother(dz,
                                                          self.kernel,
                                                          filter_mean,
                                                          filter_cov,
                                                          return_full=True,
                                                          parallel=self.parallel)

        minf, Pinf = self.minf[None, ...], self.kernel.stationary_covariance()[None, ...]
        mean_aug = np.concatenate([minf, smoother_mean, minf])
        cov_aug = np.concatenate([Pinf, smoother_cov, Pinf])
        gain = np.concatenate([np.zeros_like(gain[:1]), gain])
        # construct the joint distribution between neighbouring pairs of states
        post_mean, post_cov = vmap(build_joint, [0, None, None, None])(
            np.arange(self.num_transitions), mean_aug, cov_aug, gain
        )

        self.posterior_mean.value, self.posterior_variance.value = post_mean, post_cov

    def compute_log_lik(self, pseudo_y=None, pseudo_var=None):
        """
        Compute the log marginal likelihood of the pseudo model, i.e. the log normaliser of the approximate posterior
        """
        log_lik, (_, _) = self.filter(self.dz,
                                      self.kernel,
                                      self.pseudo_likelihood.mean,
                                      self.pseudo_likelihood.covariance,
                                      parallel=self.parallel)
        return log_lik

    def compute_kl(self):
        """
        KL divergence between the approximate posterior q(u) and the prior p(u)
        TODO: can we remove the need for this by generalising these methods?
        """
        # log int p(u) prod_n N(pseudo_y_n | u, pseudo_var_n) du
        log_lik_pseudo = self.compute_log_lik()
        # E_q[log N(pseudo_y_n | u, pseudo_var_n)]
        expected_density_pseudo = self.expected_density_pseudo()
        kl = expected_density_pseudo - log_lik_pseudo  # KL[approx_post || prior]
        return kl

    def predict(self, X, R=None):
        """
        predict at new test locations X
        """
        if len(X.shape) < 2:
            X = X[:, None]
        if R is None:
            R = X[:, 1:]
        X = X[:, :1]  # take only the temporal component

        _, (filter_mean, filter_cov) = self.filter(self.dz,
                                                   self.kernel,
                                                   self.pseudo_likelihood.mean,
                                                   self.pseudo_likelihood.covariance,
                                                   parallel=self.parallel)
        dz = self.dz[1:]
        smoother_mean, smoother_cov, gain = self.smoother(dz,
                                                          self.kernel,
                                                          filter_mean,
                                                          filter_cov,
                                                          return_full=True,
                                                          parallel=self.parallel)

        # predict the state distribution at the test time steps
        state_mean, state_cov = self.temporal_conditional(self.Z.value, X, smoother_mean, smoother_cov,
                                                          gain, self.kernel)
        # extract function values from the state:
        H = self.kernel.measurement_model()
        if self.spatio_temporal:
            # TODO: if R is fixed, only compute B, C once
            B, C = self.kernel.spatial_conditional(X, R)
            W = B @ H
            test_mean = W @ state_mean
            test_var = W @ state_cov @ transpose(W) + C
        else:
            test_mean, test_var = H @ state_mean, H @ state_cov @ transpose(H)

        if np.squeeze(test_var).ndim > 2:  # deal with spatio-temporal case (discard spatial covariance)
            test_var = diag(np.squeeze(test_var))
        return np.squeeze(test_mean), np.squeeze(test_var)

    def conditional_posterior_to_data(self, batch_ind=None, post_mean=None, post_cov=None):
        """
        compute
        q(f) = int p(f | u) q(u) du
        where
        q(u) = N(u | post_mean, post_cov)
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        if post_mean is None:
            post_mean = self.posterior_mean.value
        if post_cov is None:
            post_cov = self.posterior_variance.value
        ind = self.ind[batch_ind]
        post_mean, post_cov = post_mean[ind], post_cov[ind]

        P, T = vmap(compute_conditional_statistics, [0, None, None, 0])(
            self.X[batch_ind, :1], self.Z.value, self.kernel, ind
        )

        H = self.kernel.measurement_model()
        if self.spatio_temporal:
            B, C = self.kernel.spatial_conditional(self.X[batch_ind], self.R[batch_ind])
            BH = B @ H
            self.conditional_mean = BH @ P  # W
            conditional_cov = BH @ T @ transpose(BH) + C  # nu
        else:
            self.conditional_mean = H @ P  # W
            conditional_cov = H @ T @ transpose(H)  # nu

        mean_f = self.conditional_mean @ post_mean
        cov_f = self.conditional_mean @ post_cov @ transpose(self.conditional_mean) + conditional_cov

        return mean_f, cov_f

    def conditional_data_to_posterior(self, mean_f, cov_f):
        """
        conditional_posterior_to_data() must be run first so that self.conditional_mean is set
        """
        mean_q = transpose(self.conditional_mean) @ mean_f
        cov_q = transpose(self.conditional_mean) @ cov_f @ self.conditional_mean
        return mean_q, cov_q

    def group_natural_params(self, nat1_n, nat2_n, batch_ind=None):

        if batch_ind is None:
            ind = self.ind
        else:
            ind = self.ind[batch_ind]

        old_nat1 = self.pseudo_likelihood.nat1
        old_nat2 = self.pseudo_likelihood.nat2

        (new_nat1, new_nat2, counter), _ = scan(f=sum_natural_params_by_group,
                                                init=(np.zeros_like(old_nat1),
                                                      np.zeros_like(old_nat2),
                                                      np.zeros(old_nat1.shape[0])),
                                                xs=(ind, nat1_n, nat2_n))

        num_neighbours = np.maximum(self.num_neighbours, 1).reshape(-1, 1, 1)
        counter = counter.reshape(-1, 1, 1)
        nat1 = new_nat1 + (1. - counter / num_neighbours) * old_nat1
        nat2 = new_nat2 + (1. - counter / num_neighbours) * old_nat2

        nat2 += 1e-8 * np.eye(nat2.shape[1])  # prevent zeros

        return nat1, nat2

    def cavity_distribution(self, batch_ind=None, power=None):
        """ Compute the power EP cavity for the given data points """
        fraction = power / np.maximum(self.num_neighbours, 1)
        cavity_mean, cavity_cov = vmap(compute_cavity)(
            self.posterior_mean.value,
            self.posterior_variance.value,
            self.pseudo_likelihood.nat1,
            self.pseudo_likelihood.nat2,
            fraction
        )
        return cavity_mean, cavity_cov


SparseMarkovGP = SparseMarkovGaussianProcess


class MarkovMeanFieldGaussianProcess(MarkovGaussianProcess):
    """
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 R=None,
                 parallel=None):
        super().__init__(kernel, likelihood, X, Y, R=R, parallel=parallel)
        self.block_index = self.kernel.get_meanfield_block_index()

    def filter(self, *args, **kwargs):
        return kalman_filter_meanfield(*args, **kwargs, block_index=self.block_index)

    def smoother(self, *args, **kwargs):
        return rauch_tung_striebel_smoother_meanfield(*args, **kwargs, block_index=self.block_index)

    def predict(self, X=None, R=None, pseudo_lik_params=None):
        """
        TODO: THIS IS HERE TO HANDLE A BUG IN PREDICTIONS WHEN USING THE PARALLEL FILTER WITH MEAN-FIELD
              YET TO FIGURE OUT WHAT'S CAUSING IT - VERY STRANGE THAT IT'S ONLY AN ISSUE DURING PREDICTION
        """
        if X is None:
            X = self.X
        elif len(X.shape) < 2:
            X = X[:, None]
        if R is None:
            R = X[:, 1:]
        X = X[:, :1]  # take only the temporal component

        if pseudo_lik_params is None:
            pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        else:
            pseudo_y, pseudo_var = pseudo_lik_params  # this deals with the posterior sampling case
        _, (filter_mean, filter_cov) = self.filter(self.dt,
                                                   self.kernel,
                                                   pseudo_y,
                                                   pseudo_var,
                                                   mask=self.mask_pseudo_y,  # mask has no effect here (loglik not used)
                                                   # parallel=self.parallel)
                                                   parallel=False)  # TODO: <---------- can't use parallel here, WHY?!
        dt = np.concatenate([self.dt[1:], np.array([0.0])], axis=0)
        smoother_mean, smoother_cov, gain = self.smoother(dt,
                                                          self.kernel,
                                                          filter_mean,
                                                          filter_cov,
                                                          return_full=True,
                                                          parallel=self.parallel)

        # add dummy states at either edge
        inf = 1e10 * np.ones_like(self.X[0, :1])
        X_aug = np.block([[-inf], [self.X[:, :1]], [inf]])

        # predict the state distribution at the test time steps:
        state_mean, state_cov = self.temporal_conditional(X_aug, X, smoother_mean, smoother_cov, gain, self.kernel)
        # extract function values from the state:
        H = self.kernel.measurement_model()
        if self.spatio_temporal:
            # TODO: if R is fixed, only compute B, C once
            B, C = self.kernel.spatial_conditional(X, R)
            W = B @ H
            test_mean = W @ state_mean
            test_var = W @ state_cov @ transpose(W) + C
        else:
            test_mean, test_var = H @ state_mean, H @ state_cov @ transpose(H)

        if np.squeeze(test_var).ndim > 2:  # deal with spatio-temporal case (discard spatial covariance)
            test_var = diag(np.squeeze(test_var))
        return np.squeeze(test_mean), np.squeeze(test_var)


MarkovMeanFieldGP = MarkovMeanFieldGaussianProcess


class SparseMarkovMeanFieldGaussianProcess(SparseMarkovGaussianProcess):
    """
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 Z,
                 R=None):
        super().__init__(kernel, likelihood, X, Y, Z, R=R, parallel=False)  # TODO: parallel version
        self.block_index = self.kernel.get_meanfield_block_index()

    def filter(self, *args, **kwargs):
        return kalman_filter_pairs_meanfield(*args, **kwargs, block_index=self.block_index)

    def smoother(self, *args, **kwargs):
        return rauch_tung_striebel_smoother_meanfield(*args, **kwargs, block_index=self.block_index)


SparseMarkovMeanFieldGP = SparseMarkovMeanFieldGaussianProcess


class InfiniteHorizonGaussianProcess(MarkovGaussianProcess):
    """
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 R=None,
                 dare_iters=20,
                 parallel=None):
        super().__init__(kernel, likelihood, X, Y, R=R, parallel=parallel)
        assert np.max(np.abs(np.diff(self.dt[1:]))) < 1e-6  # check time steps are equidistant
        self.heteroscedastic = (np.any(np.isnan(self.Y))) or (not isinstance(self.likelihood, Gaussian))
        self.dare_iters = dare_iters
        # initialise the DARE solvers with the previous result
        Pinf = self.kernel.stationary_covariance()
        self.dare_init_filter = objax.StateVar(Pinf)
        self.dare_init_smoother = objax.StateVar(Pinf)

    def filter(self, *args, **kwargs):
        dare_init = self.dare_init_filter.value
        if self.spatio_temporal:  # spatio-temporal case
            B, C = self.kernel.spatial_conditional(self.X, self.R)
            tied_pseudo_nat2 = np.mean(transpose(B) @ self.pseudo_likelihood.nat2 @ B, axis=0)
        else:
            tied_pseudo_nat2 = np.mean(self.pseudo_likelihood.nat2, axis=0)
        pseudo_var_tied = inv(tied_pseudo_nat2)
        filter_output = kalman_filter_infinite_horizon(*args, **kwargs,
                                                       heteroscedastic=self.heteroscedastic,
                                                       noise_cov_tied=pseudo_var_tied,
                                                       dare_iters=self.dare_iters,
                                                       dare_init=dare_init)
        _, (_, (Pdare, _)) = filter_output  # extract DARE solution from the filter outputs
        self.dare_init_filter.value = Pdare
        return filter_output

    def smoother(self, *args, **kwargs):
        dare_init = self.dare_init_smoother.value
        means, covs, gains, dare_cov = rauch_tung_striebel_smoother_infinite_horizon(*args, **kwargs,
                                                                                     dare_iters=self.dare_iters,
                                                                                     dare_init=dare_init)
        self.dare_init_smoother.value = dare_cov
        return means, covs, gains

    @staticmethod
    def temporal_conditional(*args, **kwargs):
        """
        This method uses the infinite-horizon stationary covariance for predictions, i.e. assumes that the
        covariance is fixed even away from the data
        """
        return temporal_conditional_infinite_horizon(*args, **kwargs)


IHGP = InfiniteHorizonGaussianProcess


class SparseInfiniteHorizonGaussianProcess(SparseMarkovGaussianProcess):
    """
    """
    def __init__(self,
                 kernel,
                 likelihood,
                 X,
                 Y,
                 Z,
                 R=None,
                 dare_iters=20):
        super().__init__(kernel, likelihood, X, Y, Z, R=R, parallel=False)  # TODO: parallel version
        assert np.max(np.abs(np.diff(self.dz[1:-1]))) < 1e-4  # check inducing time steps are equidistant
        self.dare_iters = dare_iters
        # initialise the DARE solvers with the previous result
        Pinf = self.kernel.stationary_covariance()
        A = self.kernel.state_transition(self.dz[1])
        Q = process_noise_covariance(A, Pinf)
        Pinf_joint = np.block([[Pinf,     Pinf @ A.T],
                               [A @ Pinf, A @ Pinf @ A.T + Q]])
        self.dare_init_filter = objax.StateVar(Pinf_joint)
        self.dare_init_smoother = objax.StateVar(Pinf)

    def filter(self, *args, **kwargs):
        dare_init = self.dare_init_filter.value
        tied_pseudo_nat2 = np.mean(self.pseudo_likelihood.nat2, axis=0)
        pseudo_var_tied = inv(tied_pseudo_nat2)
        filter_output = kalman_filter_infinite_horizon_pairs(*args, **kwargs,
                                                             noise_cov_tied=pseudo_var_tied,
                                                             dare_iters=self.dare_iters,
                                                             dare_init=dare_init)
        _, (_, (Pdare, _)) = filter_output  # extract DARE solution from the filter outputs
        self.dare_init_filter.value = Pdare
        return filter_output

    def smoother(self, *args, **kwargs):
        dare_init = self.dare_init_smoother.value
        means, covs, gains, dare_cov = rauch_tung_striebel_smoother_infinite_horizon(*args, **kwargs,
                                                                                     dare_iters=self.dare_iters,
                                                                                     dare_init=dare_init)
        self.dare_init_smoother.value = dare_cov
        return means, covs, gains

    @staticmethod
    def temporal_conditional(*args, **kwargs):
        """
        This method uses the infinite-horizon stationary covariance for predictions, i.e. assumes that the
        covariance is fixed even away from the data
        """
        return temporal_conditional_infinite_horizon(*args, **kwargs)


SparseIHGP = SparseInfiniteHorizonGaussianProcess



import bayesnewton
import objax
import numpy as np
from jax.config import config
config.update("jax_enable_x64", True)
import pytest

# TODO: ------- FIX --------


def create_grid(x1, x2, y1, y2, n1=10, n2=10):
    y_ = np.linspace(y1, y2, n2)
    x_ = np.linspace(x1, x2, n1)

    grid = []
    for i in x_:
        for j in y_:
            grid.append([i, j])

    return np.array(grid)


def build_data(N_):
    Nt_train = N_
    Ns = N_
    X_ = create_grid(0, 1, 0, 1, Nt_train, Ns)
    t_ = np.linspace(0, 1, Nt_train, dtype=float)
    R_ = np.tile(np.linspace(0, 1, Ns, dtype=float)[None, ...], [Nt_train, 1])

    y_ = np.sin(10 * X_[:, 0]) + np.sin(10 * X_[:, 1]) + 0.01 * np.random.randn(X_.shape[0])

    # Y = y[:, None]
    Y_ = y_.reshape(Nt_train, Ns)
    return X_, Y_, t_, R_, y_


def initialise_gp_model(var_f, len_f, var_y, x_, y_, z_):
    kernel = bayesnewton.kernels.SpatialMatern52(variance=var_f, lengthscale=len_f,
                                          z=z_, sparse=True, opt_z=False, conditional='Full')
    likelihood = bayesnewton.likelihoods.Gaussian(variance=var_y)

    # the sort during utils.input_admin() sometimes results in different sorting of inputs
    # so this step ensures everything is aligned
    model_ = bayesnewton.models.MarkovVariationalGP(kernel=kernel, likelihood=likelihood, X=x_, Y=y_)
    x_sorted = model_.X
    r_sorted = model_.R
    x_ = np.vstack([x_sorted.T, r_sorted.T]).T
    y_ = model_.Y

    model = bayesnewton.models.VariationalGP(kernel=kernel, likelihood=likelihood, X=x_, Y=y_)
    return model


def initialise_markovgp_model(var_f, len_f, var_y, x_, y_, z_):
    kernel = bayesnewton.kernels.SpatialMatern52(variance=var_f, lengthscale=len_f,
                                          z=z_, sparse=True, opt_z=False, conditional='Full')
    likelihood = bayesnewton.likelihoods.Gaussian(variance=var_y)
    model = bayesnewton.models.MarkovVariationalGP(kernel=kernel, likelihood=likelihood, X=x_, Y=y_)
    return model


@pytest.mark.parametrize('var_f', [0.5, 1.5])
@pytest.mark.parametrize('len_f', [0.75, 2.5])
@pytest.mark.parametrize('var_y', [0.1, 0.5])
@pytest.mark.parametrize('N', [8, 16])
def test_initial_loss(var_f, len_f, var_y, N):
    """
    test whether VI with newt's GP and MarkovGP give the same initial ELBO and posterior
    """

    x, Y, t, R, y = build_data(N)

    gp_model = initialise_gp_model(var_f, len_f, var_y, x, y, R[0])
    markovgp_model = initialise_markovgp_model(var_f, len_f, var_y, x, y, R[0])

    gp_model.update_posterior()
    loss_gp = gp_model.energy()
    print(loss_gp)

    markovgp_model.update_posterior()
    loss_markovgp = markovgp_model.energy()
    print(loss_markovgp)

    # print(gp_model.posterior_variance.value - markovgp_model.posterior_variance.value)

    markovgp_mean, markovgp_var = markovgp_model.predict(t, R)

    # np.testing.assert_allclose(gp_model.posterior_mean.value, markovgp_model.posterior_mean.value, rtol=1e-4)
    # np.testing.assert_allclose(gp_model.posterior_mean.value, markovgp_mean.reshape(-1, 1, 1), rtol=1e-4)
    # np.testing.assert_allclose(gp_model.posterior_variance.value, markovgp_model.posterior_variance.value, rtol=1e-4)
    # np.testing.assert_allclose(gp_model.posterior_variance.value, markovgp_var.reshape(-1, 1, 1), rtol=1e-4)
    np.testing.assert_almost_equal(loss_gp, loss_markovgp, decimal=2)


@pytest.mark.parametrize('var_f', [0.5, 1.5])
@pytest.mark.parametrize('len_f', [0.75, 2.5])
@pytest.mark.parametrize('var_y', [0.1, 0.5])
@pytest.mark.parametrize('N', [8, 16])
def test_gradient_step(var_f, len_f, var_y, N):
    """
    test whether VI with newt's GP and MarkovGP provide the same initial gradient step in the hyperparameters
    """

    x, Y, t, R, y = build_data(N)

    gp_model = initialise_gp_model(var_f, len_f, var_y, x, y, R[0])
    markovgp_model = initialise_markovgp_model(var_f, len_f, var_y, x, y, R[0])

    gv = objax.GradValues(gp_model.energy, gp_model.vars())
    gv_markov = objax.GradValues(markovgp_model.energy, markovgp_model.vars())

    lr_adam = 0.1
    lr_newton = 1.
    opt = objax.optimizer.Adam(gp_model.vars())
    opt_markov = objax.optimizer.Adam(markovgp_model.vars())

    gp_model.update_posterior()
    gp_grads, gp_value = gv()
    gp_loss_ = gp_value[0]
    opt(lr_adam, gp_grads)
    gp_hypers = np.array([gp_model.kernel.temporal_kernel.lengthscale,
                          gp_model.kernel.temporal_kernel.variance,
                          gp_model.kernel.spatial_kernel.lengthscale,
                          gp_model.likelihood.variance])
    print(gp_hypers)
    print(gp_grads)

    markovgp_model.update_posterior()
    markovgp_grads, markovgp_value = gv_markov()
    markovgp_loss_ = markovgp_value[0]
    opt_markov(lr_adam, markovgp_grads)
    markovgp_hypers = np.array([markovgp_model.kernel.temporal_kernel.lengthscale,
                                markovgp_model.kernel.temporal_kernel.variance,
                                markovgp_model.kernel.spatial_kernel.lengthscale,
                                markovgp_model.likelihood.variance])

    print(markovgp_hypers)
    print(markovgp_grads)

    np.testing.assert_allclose(gp_grads[0], markovgp_grads[0], rtol=1e-4)
    np.testing.assert_allclose(gp_grads[1], markovgp_grads[1], rtol=1e-4)
    np.testing.assert_allclose(gp_grads[2], markovgp_grads[2], rtol=1e-4)


@pytest.mark.parametrize('var_f', [0.5, 1.5])
@pytest.mark.parametrize('len_f', [0.75, 2.5])
@pytest.mark.parametrize('var_y', [0.1, 0.5])
@pytest.mark.parametrize('N', [8, 16])
def test_inference_step(var_f, len_f, var_y, N):
    """
    test whether VI with newt's GP and MarkovGP give the same posterior after one natural gradient step
    """

    x, Y, t, R, y = build_data(N)

    gp_model = initialise_gp_model(var_f, len_f, var_y, x, y, R[0])
    markovgp_model = initialise_markovgp_model(var_f, len_f, var_y, x, y, R[0])

    lr_newton = 1.

    gp_model.inference(lr=lr_newton)  # update variational params

    markovgp_model.inference(lr=lr_newton)  # update variational params

    np.testing.assert_allclose(gp_model.posterior_mean.value, markovgp_model.posterior_mean.value, rtol=1e-4)
    np.testing.assert_allclose(gp_model.posterior_variance.value, markovgp_model.posterior_variance.value, rtol=1e-4)


# N = 5
# x, Y, t, R, y = build_data(N)
#
# var_f = 0.5
# len_f = 0.75
# var_y = 0.1
#
# gp_model = initialise_gp_model(var_f, len_f, var_y, x, y, R[0])
# markovgp_model = initialise_markovgp_model(var_f, len_f, var_y, x, y, R[0])
#
# lr_newton = 1.
#
# gp_model.update_posterior()
# gp_loss = inf(gp_model, lr=lr_newton)  # update variational params
# print(gp_loss)
# gp_model.update_posterior()
#
# markovgp_model.update_posterior()
# markovgp_loss = inf(markovgp_model, lr=lr_newton)  # update variational params
# print(markovgp_loss)
# markovgp_model.update_posterior()
#
# np.testing.assert_allclose(gp_model.posterior_mean.value, markovgp_model.posterior_mean.value, rtol=1e-4)
# np.testing.assert_allclose(gp_model.posterior_variance.value, markovgp_model.posterior_variance.value, rtol=1e-4)



import objax
import jax.numpy as np
from jax import vmap
from .utils import (
    diag,
    transpose,
    inv_vmap,
    vmap_diag,
    solve,
    ensure_diagonal_positive_precision,
    mvn_logpdf_
)
from .likelihoods import Likelihood, MultiLatentLikelihood
from .basemodels import GaussianDistribution, SparseGP
import math
import abc

LOG2PI = math.log(2 * math.pi)


def newton_update(mean, jacobian, hessian):
    """
    Applies one step of Newton's method to update the pseudo_likelihood parameters.
    Note that this is the natural parameter form of a Newton step.
    """

    # deal with missing data
    hessian = np.where(np.isnan(hessian), -1e-6, hessian)
    jacobian = np.where(np.isnan(jacobian), hessian @ mean, jacobian)

    # Newton update
    pseudo_likelihood_nat1 = (
        jacobian - hessian @ mean
    )
    pseudo_likelihood_nat2 = (
        -hessian
    )

    return pseudo_likelihood_nat1, pseudo_likelihood_nat2


class InferenceMixin(abc.ABC):
    """
    The approximate inference class. To be used as a Mixin, to add inference functionality to the model class.
    Each approximate inference scheme implements an 'update_()' method which is called during
    inference in order to update the local likelihood approximation (the sites).
    TODO: improve code sharing between classes
    TODO: re-derive and re-implement QuasiNewton methods
    TODO: move as much of the generic functionality as possible from the base model class to this class.
    """
    num_data: float
    Y: np.DeviceArray
    ind: np.DeviceArray
    pseudo_likelihood: GaussianDistribution
    posterior_mean: objax.StateVar
    posterior_var: objax.StateVar
    update_posterior: classmethod
    group_natural_params: classmethod
    set_pseudo_likelihood: classmethod
    conditional_posterior_to_data: classmethod
    conditional_data_to_posterior: classmethod
    likelihood: Likelihood

    def inference(self, lr=1., batch_ind=None, **kwargs):

        if (batch_ind is None) or (batch_ind.shape[0] == self.num_data):
            batch_ind = None

        self.update_posterior()  # make sure the posterior is up to date

        # use the chosen inference method (VI, EP, ...) to compute the necessary terms for the parameter update
        mean, jacobian, hessian = self.update_variational_params(batch_ind, lr, **kwargs)
        # ---- Newton update ----
        nat1_n, nat2_n = newton_update(mean, jacobian, hessian)
        # -----------------------
        nat1, nat2 = self.group_natural_params(nat1_n, nat2_n, batch_ind)  # only required for SparseMarkov models

        diff1 = np.mean(np.abs(nat1 - self.pseudo_likelihood.nat1))
        diff2 = np.mean(np.abs(nat2 - self.pseudo_likelihood.nat2))

        # ---- update the model variational parameters ----
        self.pseudo_likelihood.update_nat_params(
            nat1=(1 - lr) * self.pseudo_likelihood.nat1 + lr * nat1,
            nat2=(1 - lr) * self.pseudo_likelihood.nat2 + lr * nat2
        )

        self.update_posterior()  # recompute posterior with new params

        return (mean, jacobian, hessian), (diff1, diff2)  # output state to be used in linesearch methods

    def update_variational_params(self, batch_ind=None, lr=1., **kwargs):
        raise NotImplementedError

    def energy(self, batch_ind=None, **kwargs):
        raise NotImplementedError


class Newton(InferenceMixin):
    """
    Newton = Laplace
    """
    compute_kl: classmethod

    def update_variational_params(self, batch_ind=None, lr=1., ensure_psd=True, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, _ = self.conditional_posterior_to_data(batch_ind)

        # Laplace approximates the expected density with a point estimate at the posterior mean: log p(y|f=m)
        log_lik, jacobian, hessian = vmap(self.likelihood.log_likelihood_gradients)(  # parallel
            self.Y[batch_ind],
            mean_f
        )

        if ensure_psd:
            hessian = -ensure_diagonal_positive_precision(-hessian)  # manual fix to avoid non-PSD precision

        jacobian, hessian = self.conditional_data_to_posterior(jacobian[..., None], hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case

    def energy(self, batch_ind=None, **kwargs):
        """
        TODO: implement correct Laplace energy
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
            scale = 1
        else:
            scale = self.num_data / batch_ind.shape[0]

        mean_f, _ = self.conditional_posterior_to_data(batch_ind)

        # Laplace approximates the expected density with a point estimate at the posterior mean: log p(y|f=m)
        log_lik, _, _ = vmap(self.likelihood.log_likelihood_gradients)(  # parallel
            self.Y[batch_ind],
            mean_f
        )

        KL = self.compute_kl()  # KL[q(f)|p(f)]
        laplace_energy = -(  # Laplace approximation to the negative log marginal likelihood
            scale * np.nansum(log_lik)  # nansum accounts for missing data
            - KL
        )

        return laplace_energy


Laplace = Newton


class VariationalInference(InferenceMixin):
    """
    Natural gradient VI (using the conjugate-computation VI approach)
    Refs:
        Khan & Lin 2017 "Conugate-computation variational inference - converting inference
                         in non-conjugate models in to inference in conjugate models"
        Chang, Wilkinson, Khan & Solin 2020 "Fast variational learning in state space Gaussian process models"
    """
    compute_kl: classmethod

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, ensure_psd=True, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # VI expected density is expected log-likelihood: E_q[log p(y|f)]
        ell, dell_dm, d2ell_dm2 = vmap(self.likelihood.variational_expectation, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature
        )

        if ensure_psd:
            d2ell_dm2 = -ensure_diagonal_positive_precision(-d2ell_dm2)  # manual fix to avoid non-PSD precision

        jacobian, hessian = self.conditional_data_to_posterior(dell_dm, d2ell_dm2)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case

    def energy(self, batch_ind=None, cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
            scale = 1
        else:
            scale = self.num_data / batch_ind.shape[0]

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # VI expected density is expected log-likelihood: E_q[log p(y|f)]
        ell, _, _ = vmap(self.likelihood.variational_expectation, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature
        )

        KL = self.compute_kl()  # KL[q(f)|p(f)]
        variational_free_energy = -(  # the variational free energy, i.e., the negative ELBO
            scale * np.nansum(ell)  # nansum accounts for missing data
            - KL
        )

        return variational_free_energy


class ExpectationPropagation(InferenceMixin):
    """
    Expectation propagation (EP)
    """
    power: float
    cavity_distribution: classmethod
    cavity_distribution_tied: classmethod
    compute_full_pseudo_lik: classmethod
    compute_log_lik: classmethod
    compute_ep_energy_terms: classmethod
    mask_y: np.DeviceArray
    mask_pseudo_y: np.DeviceArray

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, ensure_psd=True, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        # compute the cavity distribution
        if isinstance(self, SparseGP):
            # we use the tied version for sparse EP because it is much more efficient
            cavity_mean, cavity_cov = self.cavity_distribution_tied(batch_ind, self.power)
        else:
            cavity_mean, cavity_cov = self.cavity_distribution(batch_ind, self.power)

        cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(batch_ind, cavity_mean, cavity_cov)

        # calculate log marginal likelihood and the new sites via moment matching:
        # EP expected density is the log expected likelihood: log E_q[p(y|f)]
        lel, dlel, d2lel = vmap(self.likelihood.moment_match, (0, 0, 0, None, None))(
            self.Y[batch_ind],
            cav_mean_f,
            cav_cov_f,
            self.power,
            cubature
        )

        cav_prec = inv_vmap(cav_cov_f)
        scale_factor = cav_prec @ inv_vmap(d2lel + cav_prec) / self.power  # this form guarantees symmetry

        dlel = scale_factor @ dlel
        d2lel = scale_factor @ d2lel
        if self.mask_pseudo_y is not None:
            # apply mask
            mask = self.mask_pseudo_y[batch_ind][..., None]
            dlel = np.where(mask, np.nan, dlel)
            d2lZ_masked = np.where(mask + transpose(mask), 0., d2lel)  # ensure masked entries are independent
            d2lel = np.where(diag(mask)[..., None], np.nan, d2lZ_masked)  # ensure masked entries return log like of 0

        if ensure_psd:
            d2lel = -ensure_diagonal_positive_precision(-d2lel)  # manual fix to avoid non-PSD precision

        jacobian, hessian = self.conditional_data_to_posterior(dlel, d2lel)

        if cav_mean_f.shape[1] == jacobian.shape[1]:
            return cav_mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return cavity_mean[ind], jacobian, hessian  # sparse Markov case

    def energy(self, batch_ind=None, cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
            scale = 1.
        else:
            scale = self.num_data / batch_ind.shape[0]

        if isinstance(self, SparseGP):
            (cavity_mean, cavity_cov), lel_pseudo, lZ = self.compute_ep_energy_terms(batch_ind, self.power)
            cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(batch_ind, cavity_mean, cavity_cov)
        else:
            # TODO: the batch indices could be handled better here
            (cavity_mean, cavity_cov), lel_pseudo, lZ = self.compute_ep_energy_terms(None, self.power)
            cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(None, cavity_mean, cavity_cov)
            cav_mean_f, cav_cov_f = cav_mean_f[batch_ind], cav_cov_f[batch_ind]

        # EP expected density is log expected likelihood: log E_q[p(y|f)]
        lel, _, _ = vmap(self.likelihood.moment_match, (0, 0, 0, None, None))(
            self.Y[batch_ind],
            cav_mean_f,
            cav_cov_f,
            self.power,
            cubature
        )

        if self.mask_y is not None:
            # TODO: this assumes masking is implemented in MultiLatentLikelihood - generalise
            if not isinstance(self.likelihood, MultiLatentLikelihood):
                if np.squeeze(self.mask_y[batch_ind]).ndim != np.squeeze(lel).ndim:
                    raise NotImplementedError('masking in spatio-temporal models not implemented for EP')
                lel = np.where(np.squeeze(self.mask_y[batch_ind]), 0., np.squeeze(lel))  # apply mask

        ep_energy = -(
            lZ
            + 1. / self.power * (scale * np.nansum(lel) - np.nansum(lel_pseudo))
        )

        return ep_energy


class PosteriorLinearisation(InferenceMixin):
    """
    An iterated smoothing algorithm based on statistical linear regression (SLR).
    This method linearises the likelihood model in the region described by the posterior.
    """
    # TODO: remove these when possible
    cavity_distribution: classmethod
    compute_full_pseudo_lik: classmethod
    mask_y: np.DeviceArray
    mask_pseudo_y: np.DeviceArray

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # PL expected density is mu=E_q[E(y|f)]
        mu, omega, d_mu, _ = vmap(self.likelihood.statistical_linear_regression, (0, 0, None))(
            mean_f,
            cov_f,
            cubature
        )
        residual = self.Y[batch_ind].reshape(mu.shape) - mu

        # deal with missing data
        mask = np.isnan(residual)
        residual = np.where(mask, 0., residual)
        omega = np.where(mask + transpose(mask), 0., omega)
        omega = np.where(vmap_diag(mask[..., 0]), 1e6, omega)

        dmu_omega = transpose(solve(omega, d_mu))  # d_mu^T inv(omega)
        jacobian = dmu_omega @ residual
        hessian = -dmu_omega @ d_mu

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case

    def energy(self, batch_ind=None, cubature=None, **kwargs):
        """
        The PL energy given in [1] is a poor approximation to the EP energy (although the gradients are ok, since
        the part they discard does not depends on the hyperparameters). Therefore, we can choose to use either
        the variational free energy or EP energy here.
        TODO: develop a PL energy approximation that reuses the linearisation quantities and matches GHS / UKS etc.
        [1] Garcia-Fernandez, Tronarp, Särkkä (2018) 'Gaussian process classification using posterior linearisation'
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
            scale = 1
        else:
            scale = self.num_data / batch_ind.shape[0]

        # compute the cavity distribution
        cavity_mean, cavity_cov = self.cavity_distribution(None, 1.)  # TODO: check batch_ind is not required
        cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(None, cavity_mean, cavity_cov)

        # calculate log marginal likelihood and the new sites via moment matching:
        # EP expected density is log E_q[p(y|f)]
        lZ, _, _ = vmap(self.likelihood.moment_match, (0, 0, 0, None, None))(
            self.Y[batch_ind],
            cav_mean_f[batch_ind],
            cav_cov_f[batch_ind],
            1.,
            cubature
        )

        if self.mask_y is not None:
            # TODO: this assumes masking is implemented in MultiLatentLikelihood - generalise
            if not isinstance(self.likelihood, MultiLatentLikelihood):
                if np.squeeze(self.mask_y[batch_ind]).ndim != np.squeeze(lZ).ndim:
                    raise NotImplementedError('masking in spatio-temporal models not implemented for EP')
                lZ = np.where(np.squeeze(self.mask_y[batch_ind]), 0., np.squeeze(lZ))  # apply mask

        pseudo_y, pseudo_var = self.compute_full_pseudo_lik()
        lZ_pseudo = mvn_logpdf_(
            pseudo_y,
            cavity_mean,
            pseudo_var + cavity_cov,
            self.mask_pseudo_y
        )

        if isinstance(self, SparseGP):
            pseudo_y, pseudo_var = self.compute_global_pseudo_lik()
            # TODO: check derivation for SparseGP. May take different form
            # lZ_post = np.sum(vmap(self.compute_log_lik)(pseudo_y, pseudo_var))
        # else:
        lZ_post = self.compute_log_lik(pseudo_y, pseudo_var)

        ep_energy = -(
                lZ_post
                + (scale * np.nansum(lZ) - np.nansum(lZ_pseudo))
        )

        return ep_energy


class PosteriorLinearisation2ndOrder(PosteriorLinearisation):
    """
    """
    # TODO: remove these when possible
    compute_full_pseudo_lik: classmethod
    mask_y: np.DeviceArray
    mask_pseudo_y: np.DeviceArray

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, ensure_psd=True, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, jacobian, hessian = vmap(self.likelihood.statistical_linear_regression_newton, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature,
        )

        if ensure_psd:
            hessian = -ensure_diagonal_positive_precision(-hessian)  # manual fix to avoid non-PSD precision

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case


class Taylor(Newton):
    """
    Inference using analytical linearisation, i.e. a first order Taylor expansion. This is equivalent to
    the Extended Kalman Smoother when using a Markov GP.

    Note: this class inherits the energy from the Laplace method. This is suitable because Taylor = a Gauss-Newton
    approximation to Laplace (which uses Newton's method). Their energy should be equivalent since the only
    difference between the algorithms is the way the Hessian of the likelihood is computed / approximated, which
    does not effect the energy.
    """

    def update_variational_params(self, batch_ind=None, lr=1., **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        Y = self.Y[batch_ind]

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # calculate the Jacobian of the observation model w.r.t. function fₙ and noise term rₙ
        Jf, _, Jsigma, _ = vmap(self.likelihood.analytical_linearisation)(mean_f, np.zeros_like(Y))  # evaluate at mean

        obs_cov = np.eye(Y.shape[1])  # observation noise scale is w.l.o.g. 1
        sigma = Jsigma @ obs_cov @ transpose(Jsigma)
        likelihood_expectation, _ = vmap(self.likelihood.conditional_moments)(mean_f)
        residual = Y.reshape(likelihood_expectation.shape) - likelihood_expectation  # residual, yₙ-E[yₙ|fₙ]

        mask = np.isnan(residual)
        residual = np.where(mask, 0., residual)

        Jf_invsigma = transpose(solve(sigma, Jf))  # Jf^T inv(sigma)
        jacobian = Jf_invsigma @ residual
        hessian = -Jf_invsigma @ Jf

        # deal with missing data
        jacobian = np.where(mask, np.nan, jacobian)
        hessian = np.where(mask * np.eye(mask.shape[1]), np.nan, hessian)

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case


ExtendedKalmanSmoother = Taylor


class GaussNewton(Newton):
    """
    Gauss-Newton
    """

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, jacobian, hessian = vmap(self.likelihood.gauss_newton, (0, 0))(
            self.Y[batch_ind],
            mean_f
        )

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case


class VariationalGaussNewton(VariationalInference):
    """
    Variational Gauss-Newton
    """

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, jacobian, hessian = vmap(self.likelihood.variational_gauss_newton, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature
        )

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case


class PosteriorLinearisation2ndOrderGaussNewton(PosteriorLinearisation):
    """
    """
    # TODO: remove these when possible
    compute_full_pseudo_lik: classmethod
    mask_y: np.DeviceArray
    mask_pseudo_y: np.DeviceArray

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, jacobian, hessian = vmap(self.likelihood.statistical_linear_regression_gauss_newton, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature,
        )

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            return mean_f, jacobian, hessian
        else:
            ind = self.ind[batch_ind]
            return self.posterior_mean.value[ind], jacobian, hessian  # sparse Markov case


class NewtonRiemann(Newton):
    """
    """

    def update_variational_params(self, batch_ind=None, lr=1., **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, _ = self.conditional_posterior_to_data(batch_ind)

        # Laplace approximates the expected density with a point estimate at the posterior mean: log p(y|f=m)
        log_lik, jacobian, hessian = vmap(self.likelihood.log_likelihood_gradients)(  # parallel
            self.Y[batch_ind],
            mean_f
        )

        jacobian, hessian = self.conditional_data_to_posterior(jacobian[..., None], hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            G = self.pseudo_likelihood.nat2 + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance @ G)
            return mean_f, jacobian, hessian_psd
        else:
            ind = self.ind[batch_ind]
            G = self.pseudo_likelihood.nat2[ind] + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance[ind] @ G)
            return self.posterior_mean.value[ind], jacobian, hessian_psd  # sparse Markov case


LaplaceRiemann = NewtonRiemann


class VariationalInferenceRiemann(VariationalInference):
    """
    Natural gradient VI (using the conjugate-computation VI approach) with PSD constraints via Riemannian gradients
    Refs:
        Lin, Schmidt & Khan 2020 "Handling the Positive-Definite Constraint in the Bayesian Learning Rule"
        Khan & Lin 2017 "Conugate-computation variational inference - converting inference
                         in non-conjugate models in to inference in conjugate models"
        Chang, Wilkinson, Khan & Solin 2020 "Fast variational learning in state space Gaussian process models"
    TODO: implement grouped update to enable sparse markov inference
    """

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # VI expected density is E_q[log p(y|f)]
        expected_density, dE_dm, d2E_dm2 = vmap(self.likelihood.variational_expectation, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature
        )

        jacobian, hessian = self.conditional_data_to_posterior(dE_dm, d2E_dm2)

        if mean_f.shape[1] == jacobian.shape[1]:
            G = self.pseudo_likelihood.nat2 + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance @ G)
            return mean_f, jacobian, hessian_psd
        else:
            ind = self.ind[batch_ind]
            G = self.pseudo_likelihood.nat2[ind] + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance[ind] @ G)
            return self.posterior_mean.value[ind], jacobian, hessian_psd  # sparse Markov case


class ExpectationPropagationRiemann(ExpectationPropagation):
    """
    Expectation propagation (EP) with PSD constraints via Riemannian gradients
    """

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        TODO: will not currently work with SparseGP because cavity_cov is a vector (SparseGP and SparseMarkovGP use different parameterisations)
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        # compute the cavity distribution
        cavity_mean, cavity_cov = self.cavity_distribution(batch_ind, self.power)
        cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(batch_ind, cavity_mean, cavity_cov)

        # calculate log marginal likelihood and the new sites via moment matching:
        # EP expected density is log E_q[p(y|f)]
        lZ, dlZ, d2lZ = vmap(self.likelihood.moment_match, (0, 0, 0, None, None))(
            self.Y[batch_ind],
            cav_mean_f,
            cav_cov_f,
            self.power,
            cubature
        )

        cav_prec = inv_vmap(cav_cov_f)
        scale_factor = cav_prec @ inv_vmap(d2lZ + cav_prec) / self.power  # this form guarantees symmetry

        dlZ = scale_factor @ dlZ
        d2lZ = scale_factor @ d2lZ
        if self.mask_pseudo_y is not None:
            # apply mask
            mask = self.mask_pseudo_y[batch_ind][..., None]
            dlZ = np.where(mask, np.nan, dlZ)
            d2lZ_masked = np.where(mask + transpose(mask), 0., d2lZ)  # ensure masked entries are independent
            d2lZ = np.where(diag(mask)[..., None], np.nan, d2lZ_masked)  # ensure masked entries return nan

        jacobian, hessian = self.conditional_data_to_posterior(dlZ, d2lZ)

        if cav_mean_f.shape[1] == jacobian.shape[1]:
            G = self.pseudo_likelihood.nat2 + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance @ G)
            return cav_mean_f, jacobian, hessian_psd
        else:
            ind = self.ind[batch_ind]
            G = self.pseudo_likelihood.nat2[ind] + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance[ind] @ G)
            return cavity_mean[ind], jacobian, hessian_psd  # sparse Markov case


class PosteriorLinearisation2ndOrderRiemann(PosteriorLinearisation2ndOrder):
    """
    """

    def update_variational_params(self, batch_ind=None, lr=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, jacobian, hessian = vmap(self.likelihood.statistical_linear_regression_newton, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature,
        )

        jacobian, hessian = self.conditional_data_to_posterior(jacobian, hessian)

        if mean_f.shape[1] == jacobian.shape[1]:
            G = self.pseudo_likelihood.nat2 + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance @ G)
            return mean_f, jacobian, hessian_psd
        else:
            ind = self.ind[batch_ind]
            G = self.pseudo_likelihood.nat2[ind] + hessian  # TODO: sort out batching here?
            hessian_psd = hessian - 0.5 * lr * (G @ self.pseudo_likelihood.covariance[ind] @ G)
            return self.posterior_mean.value[ind], jacobian, hessian_psd  # sparse Markov case


def bfgs(mean, jacobian, mean_prev, jacobian_prev, B_prev):
    """ The BFGS update is guaranteed to result in PSD updates only if sg < 0 """
    g = jacobian - jacobian_prev
    s = mean - mean_prev
    sBs = transpose(s) @ B_prev @ s
    sg = transpose(s) @ g
    B_new = (
        B_prev
        - (B_prev @ s @ transpose(s) @ B_prev) / sBs
        + (g @ transpose(g)) / sg
    )
    B_new = np.where(sg < -1e-14, B_new, B_prev)  # don't update the hessian if non-psd
    return B_new


def damped_bfgs(mean, jacobian, mean_prev, jacobian_prev, B_prev, damping=0.8):
    """ The damped BFGS update is guaranteed to result in PSD updates """
    g = jacobian - jacobian_prev
    s = mean - mean_prev
    sBs = transpose(s) @ B_prev @ s
    sg = transpose(s) @ g
    theta = np.where(sg > (1 - damping) * sBs,
                     damping * sBs / (sBs - sg),
                     np.ones_like(sBs))
    r = theta * g + (1 - theta) * B_prev @ s
    sr = transpose(s) @ r
    B_new = (
        B_prev
        - (B_prev @ s @ transpose(s) @ B_prev) / sBs
        + (r @ transpose(r)) / sr
    )
    # sr guaranteed to be positive except for at initial point, or for points where theta=1, so we still check:
    B_new = np.where(sr < -1e-14, B_new, B_prev)
    return B_new  # convert back to NSD


def damped_bfgs_modified(mean, jacobian, mean_prev, jacobian_prev, B_prev, damping=0.8):
    """
    The damped BFGS update is guaranteed to result in PSD updates
    This version uses the undamped BFGS if damping = 1
    """
    g = jacobian - jacobian_prev
    s = mean - mean_prev
    sBs = transpose(s) @ B_prev @ s
    sg = transpose(s) @ g
    theta = np.where(sg > (1 - damping) * sBs,
                     damping * sBs / (sBs - sg),
                     np.ones_like(sBs))
    theta = np.where(damping > 0.99, np.ones_like(theta), theta)
    r = theta * g + (1 - theta) * B_prev @ s
    sr = transpose(s) @ r
    B_new = (
        B_prev
        - (B_prev @ s @ transpose(s) @ B_prev) / sBs
        + (r @ transpose(r)) / sr
    )
    # sr guaranteed to be positive except for at initial point, or for points where theta=1, so we still check:
    B_new = np.where(sr < -1e-14, B_new, B_prev)
    return B_new  # convert back to NSD


class QuasiNewtonBase(abc.ABC):
    num_data: float
    func_dim: float
    fullcov: bool
    update_posterior: classmethod
    group_natural_params: classmethod
    update_variational_params: classmethod
    pseudo_likelihood: GaussianDistribution
    mean_prev: objax.StateVar
    jacobian_prev: objax.StateVar
    hessian_approx: objax.StateVar
    posterior_variance: objax.StateVar

    def inference(self, lr=1., batch_ind=None, **kwargs):

        if (batch_ind is None) or (batch_ind.shape[0] == self.num_data):
            batch_ind = None

        self.update_posterior()  # make sure the posterior is up to date

        # use the chosen inference method (VI, EP, ...) to compute the necessary terms for the parameter update
        mean, jacobian, hessian, quasi_newton_state = self.update_variational_params(batch_ind, lr, **kwargs)

        # update tracked quasi-newton params
        quasi_newton_mean, quasi_newton_jacobian, quasi_newton_hessian = quasi_newton_state
        self.mean_prev.value = quasi_newton_mean
        self.jacobian_prev.value = quasi_newton_jacobian
        self.hessian_approx.value = (1 - lr) * self.hessian_approx.value + lr * quasi_newton_hessian

        # ---- Newton update ----
        nat1_n, nat2_n = newton_update(mean, jacobian, hessian)
        # -----------------------
        nat1, nat2 = self.group_natural_params(nat1_n, nat2_n, batch_ind)  # only required for SparseMarkov models

        diff1 = np.mean(np.abs(nat1 - self.pseudo_likelihood.nat1))
        diff2 = np.mean(np.abs(nat2 - self.pseudo_likelihood.nat2))

        # ---- update the model variational parameters ----
        # for quasi-Newton, the damping is applied during the BFGS update
        self.pseudo_likelihood.update_nat_params(
            nat1=(1 - lr) * self.pseudo_likelihood.nat1 + lr * nat1,
            # nat1=nat1,
            nat2=(1 - lr) * self.pseudo_likelihood.nat2 + lr * nat2
            # nat2=nat2
        )

        self.update_posterior()  # recompute posterior with new params

        # return (mean, jacobian, hessian), quasi_newton_state  # output state to be used in linesearch methods
        return (mean, jacobian, hessian), (diff1, diff2)  # output state to be used in linesearch methods


class QuasiNewton(QuasiNewtonBase, Newton):
    """
    TODO: implement grouped update to enable sparse markov inference
    """

    def update_variational_params(self, batch_ind=None, lr=1., damping=1., **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        ind = self.ind[batch_ind]

        mean_f, _ = self.conditional_posterior_to_data(batch_ind)

        # Laplace approximates the expected density with a point estimate at the posterior mean: log p(y|f=m)
        log_lik, jacobian, _ = vmap(self.likelihood.log_likelihood_gradients)(  # parallel
            self.Y[batch_ind],
            mean_f
        )

        jacobian, _ = self.conditional_data_to_posterior(jacobian[..., None], _)

        B = damped_bfgs_modified(
            mean_f, jacobian,
            self.mean_prev.value[ind], self.jacobian_prev.value[ind],
            self.hessian_approx.value[ind],
            damping=damping
        )

        if self.mean_prev.value.shape[0] != mean_f.shape[0]:
            B = self.hessian_approx.value.at[ind].set(B)
            jacobian = self.jacobian_prev.value.at[ind].set(jacobian) 
            mean_f = self.mean_prev.value.at[ind].set(mean_f)

        return mean_f, jacobian, B, (mean_f, jacobian, B)


class VariationalQuasiNewton(QuasiNewtonBase, VariationalInference):
    """
    TODO: implement grouped update to enable sparse markov inference
    """

    def update_variational_params(self, batch_ind=None, lr=1., damping=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        ind = self.ind[batch_ind]

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        # VI expected density is E_q[log p(y|f)]
        expected_density, dE_dm, d2E_dm2 = vmap(self.likelihood.variational_expectation, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature
        )

        dE_dv = 0.5 * d2E_dm2

        jacobian, _ = self.conditional_data_to_posterior(dE_dm, d2E_dm2)

        if self.fullcov:
            mean_var = np.concatenate([mean_f, cov_f.reshape(self.num_data, -1, 1)], axis=1)
            jacobian_mean_var = np.concatenate([dE_dm, dE_dv.reshape(self.num_data, -1, 1)], axis=1)
        else:
            mean_var = np.concatenate([mean_f, diag(cov_f)[..., None]], axis=1)
            jacobian_mean_var = np.concatenate([dE_dm, diag(dE_dv)[..., None]], axis=1)

        B = damped_bfgs_modified(
            mean_var, jacobian_mean_var,
            self.mean_prev.value[ind], self.jacobian_prev.value[ind],
            self.hessian_approx.value[ind],
            damping=damping
        )

        if self.mean_prev.value.shape[0] != mean_f.shape[0]:
            B = self.hessian_approx.value.at[ind].set(B)
            jacobian = self.jacobian_prev.value.at[ind].set(jacobian)
            mean_f = self.mean_prev.value.at[ind].set(mean_f)

        return mean_f, jacobian, B[:, :self.func_dim, :self.func_dim], (mean_var, jacobian_mean_var, B)


class ExpectationPropagationQuasiNewton(QuasiNewtonBase, ExpectationPropagation):
    """
    TODO: implement grouped update to enable sparse markov inference
    """

    def update_variational_params(self, batch_ind=None, lr=1., damping=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        ind = self.ind[batch_ind]

        # compute the cavity distribution
        cavity_mean, cavity_cov = self.cavity_distribution(batch_ind, self.power)
        cav_mean_f, cav_cov_f = self.conditional_posterior_to_data(batch_ind, cavity_mean, cavity_cov)

        # calculate log marginal likelihood and the new sites via moment matching:
        # EP expected density is log E_q[p(y|f)]
        lZ, dlZ_dm, d2lZ_dm, dlZ_dv = vmap(self.likelihood.moment_match_dv, (0, 0, 0, None, None))(
            self.Y[batch_ind],
            cav_mean_f,
            cav_cov_f,
            self.power,
            cubature
        )

        jacobian_unscaled, _ = self.conditional_data_to_posterior(dlZ_dm, d2lZ_dm)
        if self.fullcov:
            cavity_mean_var = np.concatenate([cav_mean_f, cav_cov_f.reshape(self.num_data, -1, 1)], axis=1)
            jacobian_unscaled_mean_var = np.concatenate([dlZ_dm, dlZ_dv.reshape(self.num_data, -1, 1)], axis=1)
        else:
            cavity_mean_var = np.concatenate([cav_mean_f, diag(cav_cov_f)[..., None]], axis=1)
            jacobian_unscaled_mean_var = np.concatenate([dlZ_dm, diag(dlZ_dv)[..., None]], axis=1)

        B = damped_bfgs_modified(
            cavity_mean_var, jacobian_unscaled_mean_var,
            self.mean_prev.value[ind], self.jacobian_prev.value[ind],
            self.hessian_approx.value[ind],
            damping=damping
        )

        cav_prec = inv_vmap(cav_cov_f)
        scale_factor = cav_prec @ inv_vmap(-B[:, :self.func_dim, :self.func_dim] + cav_prec) / self.power

        jacobian = scale_factor @ jacobian_unscaled
        hessian = scale_factor @ B[:, :self.func_dim, :self.func_dim]

        if self.mean_prev.value.shape[0] != cav_mean_f.shape[0]:
            B = self.hessian_approx.value.at[ind].set(B)
            jacobian = self.jacobian_prev.value.at[ind].set(jacobian) 
            cav_mean_f = self.mean_prev.value.at[ind].set(cav_mean_f)


        return cav_mean_f, jacobian, hessian, (cavity_mean_var, jacobian_unscaled_mean_var, B)


# class PosteriorLinearisationQuasiNewton(QuasiNewtonBase, PosteriorLinearisation):
#     """
#     Quasi-Newton (BFGS) Posterior Linearisation (PL)
#     TODO: implement grouped update to enable sparse markov inference
#     """
#
#     def update_variational_params(self, batch_ind=None, lr=1., damping=1., cubature=None, **kwargs):
#         """
#         """
#         if batch_ind is None:
#             batch_ind = np.arange(self.num_data)
#         ind = self.ind[batch_ind]
#
#         mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)
#
#         mu, omega, dmu_dm, d2mu_dm2 = vmap(self.likelihood.statistical_linear_regression, (0, 0, None))(
#             mean_f,
#             cov_f,
#             cubature
#         )
#         dmu_dv = 0.5 * d2mu_dm2
#         dmu_dv = dmu_dv.reshape(dmu_dv.shape[0], dmu_dv.shape[1], -1)
#
#         residual = self.Y[batch_ind].reshape(mu.shape) - mu
#
#         # deal with missing data
#         mask = np.isnan(residual)
#         residual = np.where(mask, 0., residual)
#         omega = np.where(mask + transpose(mask), 0., omega)
#         omega = np.where(vmap_diag(mask[..., 0]), 1e6, omega)
#
#         dmu_omega = transpose(solve(omega, dmu_dm))  # dmu_dm^T @ inv(omega)
#         jacobian = dmu_omega @ residual
#         hessian_not_used = -dmu_omega @ dmu_dm
#         jacobian_var = transpose(solve(omega, dmu_dv)) @ residual
#
#         jacobian, _ = self.conditional_data_to_posterior(jacobian, hessian_not_used)
#
#         if self.fullcov:
#             mean_var = np.concatenate([mean_f, cov_f.reshape(self.num_data, -1, 1)], axis=1)
#             jacobian_mean_var = np.concatenate([
#                 jacobian.reshape(self.num_data, -1, 1),
#                 jacobian_var.reshape(self.num_data, -1, 1)
#             ], axis=1)
#         else:
#             mean_var = np.concatenate([mean_f, diag(cov_f)[..., None]], axis=1)
#             d = mean_f.shape[1]
#             jacobian_mean_var = np.concatenate([
#                 jacobian,
#                 diag(jacobian_var.reshape(self.num_data, d, d))[..., None]
#             ], axis=1)
#
#         B = damped_bfgs_modified(
#             mean_var, jacobian_mean_var,
#             self.mean_prev.value[ind], self.jacobian_prev.value[ind],
#             self.hessian_approx.value[ind],
#             damping=damping
#         )
#
#         if self.mean_prev.value.shape[0] != mean_f.shape[0]:
#             B = index_update(self.hessian_approx.value, index[ind], B)
#             jacobian = index_update(self.jacobian_prev.value, index[ind], jacobian)
#             mean_f = index_update(self.mean_prev.value, index[ind], mean_f)
#
#         return mean_f, jacobian, B[:, :self.func_dim, :self.func_dim], (mean_var, jacobian_mean_var, B)


class PosteriorLinearisation2ndOrderQuasiNewton(QuasiNewtonBase, PosteriorLinearisation):
    """
    """

    def update_variational_params(self, batch_ind=None, lr=1., damping=1., cubature=None, **kwargs):
        """
        """
        if batch_ind is None:
            batch_ind = np.arange(self.num_data)
        ind = self.ind[batch_ind]

        mean_f, cov_f = self.conditional_posterior_to_data(batch_ind)

        log_target, dmu_dm, d2mu_dm2 = vmap(self.likelihood.statistical_linear_regression_quasi_newton, (0, 0, 0, None))(
            self.Y[batch_ind],
            mean_f,
            cov_f,
            cubature,
        )
        dmu_dv = 0.5 * d2mu_dm2

        jacobian, _ = self.conditional_data_to_posterior(dmu_dm, dmu_dv)

        if self.fullcov:
            mean_var = np.concatenate([mean_f, cov_f.reshape(self.num_data, -1, 1)], axis=1)
            jacobian_mean_var = np.concatenate([dmu_dm, dmu_dv.reshape(self.num_data, -1, 1)], axis=1)
        else:
            mean_var = np.concatenate([mean_f, diag(cov_f)[..., None]], axis=1)
            jacobian_mean_var = np.concatenate([dmu_dm, diag(dmu_dv)[..., None]], axis=1)

        B = damped_bfgs_modified(
            mean_var, jacobian_mean_var,
            self.mean_prev.value[ind], self.jacobian_prev.value[ind],
            self.hessian_approx.value[ind],
            damping=damping
        )

        if self.mean_prev.value.shape[0] != mean_f.shape[0]:
            B = self.hessian_approx.value.at[ind].set(B)
            jacobian = self.jacobian_prev.value.at[ind].set(jacobian) 
            mean_f = self.mean_prev.value.at[ind].set(mean_f)

        return mean_f, jacobian, B[:, :self.func_dim, :self.func_dim], (mean_var, jacobian_mean_var, B)



import bayesnewton
import objax
from bayesnewton.utils import inv
import numpy as np
from jax.config import config
config.update("jax_enable_x64", True)
import pytest
import tensorflow as tf
import gpflow
import scipy as sp
import pickle

gpflow.config.set_default_jitter(1e-20)

train_data = pickle.load(open(f'../experiments/shutters/data/train_data_0_0.pickle', "rb"))
pred_data = pickle.load(open(f'../experiments/shutters/data/pred_data_0_0.pickle', "rb"))
pred_data = pred_data['grid']

X = train_data['X']
Y = train_data['Y']
X_test = pred_data['X']
Y_test = pred_data['Y']


def initialise_newt_model(var_f, len_f, var_y, x, y):
    r = np.unique(x[:, 1])
    print(r)
    kernel = bayesnewton.kernels.SpatioTemporalMatern52(variance=var_f, lengthscale_time=len_f, lengthscale_space=len_f, z=r)
    likelihood = bayesnewton.likelihoods.Gaussian(variance=var_y)
    model = bayesnewton.models.VariationalGP(kernel=kernel, likelihood=likelihood, X=x, Y=y)
    return model


# def init_as_cvi(kern, Z_all):
#     M = Z_all.shape[0]
#
#     Kzz = kern(Z_all, Z_all)
#
#     # def inv(K):
#     #     K_chol = sp.linalg.cholesky(K+1e-3*np.eye(M), lower=True)
#     #     return sp.linalg.cho_solve((K_chol, True), np.eye(K.shape[0]))
#
#     #manual q(u) decompositin
#     nat1 = np.zeros([M, 1])
#     nat2 = inv(Kzz)
#
#     lam1 =  np.zeros([M, 1])
#     lam2 =  1e-2*np.eye(M)
#
#     # S = inv(-2*(nat2+lam2))
#     S = inv(nat2+lam2)
#     m = S @ (lam1 + nat1)
#
#     S_chol = sp.linalg.cholesky(S+1e-8*np.eye(M), lower=True)
#     S_flattened = S_chol[np.tril_indices(M, 0)]
#
#     q_mu = m
#     q_sqrt = np.array([S_chol])
#     return q_mu, q_sqrt


def initialise_gpflow_model(var_f, len_f, var_y, x, y):
    N = x.shape[0]
    k0 = gpflow.kernels.Matern52(lengthscales=[len_f], variance=var_f, active_dims=[0], name='matern1')
    k1 = gpflow.kernels.Matern52(lengthscales=[len_f], variance=1., active_dims=[1], name='matern2')
    k = k0 * k1

    # find the m and S that correspond to the same natural parameters used by CVI
    K_xx = np.array(k(x, x))
    K_xx_inv = inv(K_xx)

    print(x.shape)

    S = inv(K_xx_inv + 1e-2 * np.eye(N))
    S_chol = np.linalg.cholesky(S)
    S_chol_init = np.array([S_chol])
    # S_chol_flattened_init = np.array(S_chol[np.tril_indices(N, 0)])

    lambda_init = np.zeros((N, 1))
    m_init = S @ lambda_init

    lik = gpflow.likelihoods.Gaussian(variance=var_y)

    # data = (x, y)
    # print(x)

    model = gpflow.models.SVGP(
        inducing_variable=x,
        whiten=False,
        kernel=k,
        mean_function=None,
        likelihood=lik,
        q_mu=m_init,
        q_sqrt=S_chol_init
    )
    gpflow.utilities.set_trainable(model.inducing_variable.Z, False)
    gpflow.utilities.set_trainable(model.q_mu, False)
    gpflow.utilities.set_trainable(model.q_sqrt, False)
    return model


@pytest.mark.parametrize('var_f', [1., 5.])
@pytest.mark.parametrize('len_f', [0.1, 0.025])
@pytest.mark.parametrize('var_y', [0.1, 0.3])
def test_initial_loss(var_f, len_f, var_y):
    """
    test whether newt's VI and gpflow's SVGP (Z=X) give the same initial ELBO and posterior
    """

    newt_model = initialise_newt_model(var_f, len_f, var_y, X, Y)
    gpflow_model = initialise_gpflow_model(var_f, len_f, var_y, X, Y)

    newt_model.update_posterior()
    loss_newt = newt_model.energy()
    # _, _, expected_density = newt_model.inference(newt_model)
    print(loss_newt)
    # print(expected_density)

    data = (X, Y)
    f_mean, f_var = gpflow_model.predict_f(X)
    var_exp = np.sum(gpflow_model.likelihood.variational_expectations(f_mean, f_var, Y))
    loss_gpflow = -gpflow_model.elbo(data)
    print(loss_gpflow.numpy())
    # print(var_exp)

    # print(posterior_mean - f_mean[:, 0])

    np.testing.assert_allclose(np.squeeze(newt_model.posterior_mean.value), f_mean[:, 0], rtol=1e-4)
    np.testing.assert_allclose(np.squeeze(newt_model.posterior_variance.value), f_var[:, 0], rtol=1e-4)
    np.testing.assert_almost_equal(loss_newt, loss_gpflow.numpy(), decimal=2)


@pytest.mark.parametrize('var_f', [1., 5.])
@pytest.mark.parametrize('len_f', [0.1, 0.025])
@pytest.mark.parametrize('var_y', [0.1, 0.3])
def test_gradient_step(var_f, len_f, var_y):
    """
    test whether newt's VI and gpflow's SVGP (Z=X) provide the same initial gradient step in the hyperparameters
    """

    # x, y = build_data(N)

    newt_model = initialise_newt_model(var_f, len_f, var_y, X, Y)
    gpflow_model = initialise_gpflow_model(var_f, len_f, var_y, X, Y)

    gv = objax.GradValues(newt_model.energy, newt_model.vars())

    lr_adam = 0.1
    lr_newton = 1.
    opt = objax.optimizer.Adam(newt_model.vars())

    newt_model.update_posterior()
    newt_grads, value = gv()  # , lr=lr_newton)
    loss_ = value[0]
    opt(lr_adam, newt_grads)
    newt_hypers = np.array([newt_model.kernel.temporal_lengthscale, newt_model.kernel.spatial_lengthscale,
                            newt_model.kernel.variance, newt_model.likelihood.variance])
    print(newt_hypers)
    print(newt_grads)

    adam_opt = tf.optimizers.Adam(lr_adam)
    data = (X, Y)
    with tf.GradientTape() as tape:
        loss = -gpflow_model.elbo(data)
    _vars = gpflow_model.trainable_variables
    gpflow_grads = tape.gradient(loss, _vars)

    loss_fn = gpflow_model.training_loss_closure(data)
    adam_vars = gpflow_model.trainable_variables
    adam_opt.minimize(loss_fn, adam_vars)
    #gpflow_hypers = np.array([gpflow_model.kernel.lengthscales.numpy()[0],
    #                          gpflow_model.kernel.lengthscales.numpy()[1],
    #                          gpflow_model.kernel.variance.numpy(),
    #                          gpflow_model.likelihood.variance.numpy()])
    gpflow_hypers = np.array([gpflow_model.kernel.parameters[0].numpy(),
                              gpflow_model.kernel.parameters[2].numpy(),
                              gpflow_model.kernel.parameters[1].numpy(),
                              gpflow_model.likelihood.variance.numpy()])
    print(gpflow_hypers)
    print(gpflow_grads)

    np.testing.assert_allclose(newt_grads[0], gpflow_grads[0], atol=1e-2)  # use atol since values are so small
    np.testing.assert_allclose(newt_grads[1], gpflow_grads[1], rtol=1e-2)
    np.testing.assert_allclose(newt_grads[2], gpflow_grads[2], rtol=1e-2)


# @pytest.mark.parametrize('var_f', [0.5, 1.5])
# @pytest.mark.parametrize('len_f', [0.75, 2.5])
# @pytest.mark.parametrize('var_y', [0.1, 0.5])
# def test_inference_step(var_f, len_f, var_y):
#     """
#     test whether newt's VI and gpflow's SVGP (Z=X) give the same posterior after one natural gradient step
#     """
#
#     # x, y = build_data(N)
#
#     newt_model = initialise_newt_model(var_f, len_f, var_y, X, Y)
#     gpflow_model = initialise_gpflow_model(var_f, len_f, var_y, X, Y)
#
#     lr_newton = 1.
#
#     newt_model.update_posterior()
#     newt_loss = inf(newt_model, lr=lr_newton)  # update variational params
#     newt_model.update_posterior()
#
#     data = (X, Y[:, None)
#     with tf.GradientTape() as tape:
#         loss = -gpflow_model.elbo(data)
#
#     variational_vars = [(gpflow_model.q_mu, gpflow_model.q_sqrt)]
#     natgrad_opt = gpflow.optimizers.NaturalGradient(gamma=lr_newton)
#     loss_fn = gpflow_model.training_loss_closure(data)
#     natgrad_opt.minimize(loss_fn, variational_vars)
#
#     f_mean, f_var = gpflow_model.predict_f(X)
#
#     # print(post_mean_)
#     # print(f_mean[:, 0])
#
#     np.testing.assert_allclose(np.squeeze(newt_model.posterior_mean.value), f_mean[:, 0], rtol=1e-3)
#     np.testing.assert_allclose(np.squeeze(newt_model.posterior_variance.value), f_var[:, 0], rtol=1e-3)

var_f = 1
len_f = 1
var_y = 0.1

newt_model = initialise_newt_model(var_f, len_f, var_y, X, Y)
gpflow_model = initialise_gpflow_model(var_f, len_f, var_y, X, Y)

newt_model.update_posterior()
loss_newt = newt_model.energy()
# _, _, expected_density = newt_model.inference(newt_model)
print(loss_newt)
# print(expected_density)

data = (X, Y)
f_mean, f_var = gpflow_model.predict_f(X)
var_exp = np.sum(gpflow_model.likelihood.variational_expectations(f_mean, f_var, Y))
loss_gpflow = -gpflow_model.elbo(data)
print(loss_gpflow.numpy())
# print(var_exp)

# print(posterior_mean - f_mean[:, 0])

# np.testing.assert_allclose(np.squeeze(newt_model.posterior_mean.value), f_mean[:, 0], rtol=1e-4)
# np.testing.assert_allclose(np.squeeze(newt_model.posterior_variance.value), f_var[:, 0], rtol=1e-4)
# np.testing.assert_almost_equal(loss_newt, loss_gpflow.numpy(), decimal=2)
